import sys
import os
import pandas as pd

sys.path.append("../src/")
import pytest
import src.pypontem.utils.unit_conversion as units


def test_pressure_conversions():
    p = units.UnitConversion.Pressure(500, "psia")
    p_pa = p.convert(to_unit="pa")
    p_psig = p.convert(to_unit="psig")
    p_bar = p.convert(to_unit="bar")
    p_barg = p.convert(to_unit="barg")
    assert round(p_pa, 1) == pytest.approx(3447378.6)
    assert round(p_psig, 1) == pytest.approx(485.3)
    assert round(p_bar, 2) == pytest.approx(34.47)
    assert round(p_barg, 2) == pytest.approx(33.46)

    # p_pa = units.UnitConversion.Pressure(3447378.6, "pa")
    # p_psia = p_pa.convert(to_unit="psia")
    # p_bar = p_pa.convert(to_unit="barg")

    # assert round(p_psia, 1) == pytest.approx(500)
    # assert round(p_bar, 2) == pytest.approx(34.47)


def test_flowrate_conversions():
    q = units.UnitConversion.LiquidFlowRate(1000, "m3_s")
    q_bbld = q.convert(to_unit="bbl_d")
    q_ft3s = q.convert(to_unit="ft3_s")
    q_bblh = q.convert(to_unit="bbl_h")
    q_gpm = q.convert(to_unit="gpm")
    q_lm = q.convert(to_unit="l_min")

    assert round(q_bbld, 1) == pytest.approx(543439650.5)
    assert round(q_ft3s, 1) == pytest.approx(35314.7)
    assert round(q_bblh, 1) == pytest.approx(22643318.7)
    assert round(q_gpm, 1) == pytest.approx(15850323.1)
    assert round(q_lm, 1) == pytest.approx(59999999.9)

    q_bbld = units.UnitConversion.LiquidFlowRate(543439650.5, "bbl_d")
    q_m3s = q_bbld.convert(to_unit="m3_s")
    q_ft3s = q_bbld.convert(to_unit="ft3_s")
    q_bblh = q_bbld.convert(to_unit="bbl_h")
    q_gpm = q_bbld.convert(to_unit="gpm")
    q_lm = q_bbld.convert(to_unit="l_min")

    assert round(q_m3s, 1) == pytest.approx(1000)
    assert round(q_ft3s, 1) == pytest.approx(35314.7)
    assert round(q_bblh, 1) == pytest.approx(22643318.7)
    assert round(q_gpm, 1) == pytest.approx(15850323.1)
    assert round(q_lm, 1) == pytest.approx(59999999.9)


def test_viscosity_conversions():
    mu = units.UnitConversion.DynamicViscosity(100, "cp")
    mu_pas = mu.convert(to_unit="pas")
    mu_lbfth = mu.convert(to_unit="lb_fth")
    mu_kgmh = mu.convert(to_unit="kg_mh")
    mu_nsm2 = mu.convert(to_unit="ns_m2")
    mu_kgms = mu.convert(to_unit="kg_ms")
    assert round(mu_pas, 1) == pytest.approx(0.1)
    assert round(mu_lbfth, 1) == pytest.approx(241.9)
    assert round(mu_kgmh) == pytest.approx(360)
    assert round(mu_nsm2, 1) == pytest.approx(0.1)
    assert round(mu_kgms, 1) == pytest.approx(0.1)

    mu_pas = units.UnitConversion.DynamicViscosity(0.1, "pas")
    mu_cp = mu_pas.convert(to_unit="cp")
    mu_lbfth = mu_pas.convert(to_unit="lb_fth")
    mu_kgmh = mu_pas.convert(to_unit="kg_mh")
    mu_nsm2 = mu_pas.convert(to_unit="ns_m2")

    assert round(mu_cp, 1) == pytest.approx(100)
    assert round(mu_lbfth, 1) == pytest.approx(241.9)
    assert round(mu_kgmh) == pytest.approx(360)
    assert round(mu_nsm2, 1) == pytest.approx(0.1)


def test_area_conversions():
    a_m2 = units.UnitConversion.Area(100, "m2")
    a_cm2 = a_m2.convert(to_unit="cm2")
    a_ft2 = a_m2.convert(to_unit="ft2")
    a_in2 = a_m2.convert(to_unit="in2")
    a_ft2 = a_m2.convert(to_unit="ft2")
    a_in2 = a_m2.convert(to_unit="in2")

    assert round(a_cm2, 1) == pytest.approx(1000000)
    assert round(a_ft2, 1) == pytest.approx(1076.4)
    assert round(a_in2, 1) == pytest.approx(155000.3)
    assert round(a_ft2, 1) == pytest.approx(1076.4)
    assert round(a_in2, 1) == pytest.approx(155000.3)


def test_density_conversions():
    rho_gcc = units.UnitConversion.Density(1, "g_cm3")
    rho_kgm3 = rho_gcc.convert(to_unit="kg_m3")
    rho_lbft3 = rho_gcc.convert(to_unit="lb_ft3")
    rho_lbin3 = rho_gcc.convert(to_unit="lb_in3")

    assert round(rho_kgm3, 1) == pytest.approx(1000)
    assert round(rho_lbft3, 1) == pytest.approx(62.4)
    assert round(rho_lbin3, 3) == pytest.approx(0.036)

    rho_kgm3 = units.UnitConversion.Density(1000, "kg_m3")
    rho_gcc = rho_kgm3.convert(to_unit="g_cm3")
    rho_lbft3 = rho_kgm3.convert(to_unit="lb_ft3")
    rho_lbin3 = rho_kgm3.convert(to_unit="lb_in3")

    assert round(rho_gcc, 1) == pytest.approx(1)
    assert round(rho_lbft3, 1) == pytest.approx(62.4)
    assert round(rho_lbin3, 3) == pytest.approx(0.036)


def test_corrosionrate_conversions():
    cr_mpy = units.UnitConversion.CorrosionRate(1, "mpy")
    cr_mmy = cr_mpy.convert(to_unit="mm_y")

    assert round(cr_mmy, 3) == pytest.approx(0.025)


def test_gasstdrate_conversions():
    qg_mmscfd = units.UnitConversion.GasStandardFlowrate(1, "mmscf_d")
    qg_scfd = qg_mmscfd.convert(to_unit="scf_d")
    qg_mscfd = qg_mmscfd.convert(to_unit="mscf_d")
    qg_sm3d = qg_mmscfd.convert(to_unit="sm3_d")
    qg_sm3s = qg_mmscfd.convert(to_unit="sm3_s")
    qg_msm3d = qg_mmscfd.convert(to_unit="msm3_d")

    assert round(qg_scfd, 1) == pytest.approx(1000000.0)
    assert round(qg_mscfd, 1) == pytest.approx(1000)
    assert round(qg_sm3d, 1) == pytest.approx(28316.8)
    assert round(qg_sm3s, 2) == pytest.approx(0.33)
    assert round(qg_msm3d, 3) == pytest.approx(0.028)

    # qg_mscfd = units.UnitConversion.GasStandardFlowrate(1000, "Mscf_d")
    # qg_mmscfd = qg_mscfd.convert(to_unit="MMscf_d")

    # assert round(qg_mmscfd) == pytest.approx(1)


def test_gor_conversions():
    gor_scfstb = units.UnitConversion.GOR(300, "scf_sbbl")
    gor_mmscfbbl = gor_scfstb.convert(to_unit="mmscf_sbbl")
    gor_sm3sm3 = gor_scfstb.convert(to_unit="sm3_sm3")
    gor_scfscf = gor_scfstb.convert(to_unit="scf_scf")

    assert round(gor_mmscfbbl, 4) == pytest.approx(0.0003)
    assert round(gor_sm3sm3, 1) == pytest.approx(53.4)
    assert round(gor_scfscf, 1) == pytest.approx(53.4)


def test_kinematicvisc_conversions():
    muk_cst = units.UnitConversion.KinematicViscosity(100, "cSt")
    muk_ft2h = muk_cst.convert(to_unit="ft2_h")
    muk_m2s = muk_cst.convert(to_unit="m2_s")
    muk_in2s = muk_cst.convert(to_unit="in2_s")

    assert round(muk_ft2h, 2) == pytest.approx(3.88)
    assert round(muk_m2s, 4) == pytest.approx(0.0001)
    assert round(muk_in2s, 3) == pytest.approx(0.155)


def test_massflowrate_conversions():
    m_kgs = units.UnitConversion.MassFlowrate(1, "kg_s")
    m_lbh = m_kgs.convert(to_unit="lb_h")
    m_td = m_kgs.convert(to_unit="t_d")
    m_lbs = m_kgs.convert(to_unit="lb_s")
    m_gs = m_kgs.convert(to_unit="g_s")
    m_ktd = m_kgs.convert(to_unit="kt_d")

    assert round(m_lbh, 2) == pytest.approx(7936.64)
    assert round(m_td, 1) == pytest.approx(86.4)
    assert round(m_lbs, 1) == pytest.approx(2.2)
    assert round(m_gs, 1) == pytest.approx(1000)
    assert round(m_ktd, 3) == pytest.approx(0.086)


def test_volume_conversions():
    m_bbl = units.UnitConversion.Volume(1000, "bbl")
    m_l = m_bbl.convert(to_unit="L")
    m_m3 = m_bbl.convert(to_unit="m3")
    m_ft3 = m_bbl.convert(to_unit="ft3")
    m_gal = m_bbl.convert(to_unit="gal")
    m_cm3 = m_bbl.convert(to_unit="cm3")
    m_mmft3 = m_bbl.convert(to_unit="mmft3")
    m_ml = m_bbl.convert(to_unit="mL")

    assert round(m_l, 1) == pytest.approx(158987.3)
    assert round(m_m3, 3) == pytest.approx(158.987)
    assert round(m_ft3, 1) == pytest.approx(5614.6)
    assert round(m_gal, 1) == pytest.approx(42000)
    assert round(m_cm3, 1) == pytest.approx(158987294.9)
    assert round(m_mmft3, 4) == pytest.approx(0.0056)
    assert round(m_ml, 1) == pytest.approx(158987294.9)


def test_conductivity_conversions():
    k_wmk = units.UnitConversion.ThermalConductivity(1000, "w_mk")
    k_wmc = k_wmk.convert(to_unit="w_mc")
    k_calmhk = k_wmk.convert(to_unit="cal_mhk")
    k_btufthR = k_wmk.convert(to_unit="btu_fthr")

    assert round(k_wmc, 1) == pytest.approx(1000)
    assert round(k_calmhk, 1) == pytest.approx(860420.65)
    assert round(k_btufthR, 2) == pytest.approx(578.18)


def test_deltatemp_conversions():
    dt_c = units.UnitConversion.TemperatureDifference(12.3, "delta_degC")
    dt_f = dt_c.convert(to_unit="delta_degF")

    assert round(dt_f, 2) == pytest.approx(22.14)


def test_cp_conversions():
    cp_jkgc = units.UnitConversion.SpecificHeatCapacity(100, "j_kgc")
    cp_btulbf = cp_jkgc.convert(to_unit="btu_lbf")
    cp_btulbr = cp_jkgc.convert(to_unit="btu_lbr")
    cp_jkgk = cp_jkgc.convert(to_unit="j_kgk")

    assert round(cp_btulbf, 2) == pytest.approx(0.02)
    assert round(cp_btulbr, 2) == pytest.approx(0.02)
    assert round(cp_jkgk, 1) == pytest.approx(100)


def test_energy_conversions():
    en_btu = units.UnitConversion.Energy(1, "btu")
    en_cal = en_btu.convert(to_unit="cal")
    en_j = en_btu.convert(to_unit="j")

    assert round(en_cal, 2) == pytest.approx(252.21)
    assert round(en_j, 1) == pytest.approx(1054.6)


def test_surfacetension_conversions():
    st_nm = units.UnitConversion.SurfaceTension(1, "n_m")
    st_dyncm = st_nm.convert(to_unit="dyn_cm")
    st_mnm = st_nm.convert(to_unit="mn_m")

    assert round(st_dyncm, 1) == pytest.approx(1000.0)
    assert round(st_mnm, 1) == pytest.approx(1000.0)


def test_molecularweight_conversions():
    mw_gmol = units.UnitConversion.MolecularWeight(44.01, "g_mol")
    mw_kgmol = mw_gmol.convert(to_unit="kg_mol")
    mw_kgkmol = mw_gmol.convert(to_unit="kg_kmol")
    mw_lblbmol = mw_gmol.convert(to_unit="lb_lbmol")

    assert round(mw_kgmol, 3) == pytest.approx(0.044)
    assert round(mw_kgkmol, 2) == pytest.approx(44.01)
    assert round(mw_lblbmol, 2) == pytest.approx(44.01)


def test_molarvol_conversions():
    mv_m3mol = units.UnitConversion.MolarVolume(1, "m3_mol")
    mv_cm3mol = mv_m3mol.convert(to_unit="cm3_mol")
    mv_lmol = mv_m3mol.convert(to_unit="l_mol")
    mv_ft3lbmol = mv_m3mol.convert(to_unit="ft3_lbmol")

    assert round(mv_cm3mol, 1) == pytest.approx(1000000)
    assert round(mv_lmol, 1) == pytest.approx(1000)
    assert round(mv_ft3lbmol, 2) == pytest.approx(16018.46)



def pytest_addoption(parser):
    """
    Function to add command line arguments
    """
    parser.addoption(
        "--filepath",
        action="store",
        default="C:/Users/HP/Desktop/Pontem Analytics/tpl/new tpl/generic_test.tpl"
    )
    parser.addoption("--variable", action="store", default="PT")
    parser.addoption("--branch", action="store")
    parser.addoption("--pipe", action="store")
    parser.addoption("--index", action="store", type=int, help="Index to compare")
    parser.addoption("--value", action="store")


import pint
import math

ureg = pint.UnitRegistry()
Q_ = ureg.Quantity

unit_map = {
    "S": "second",
    "Min": "minute",
    "H": "hour",
    "D": "day",
    "W": "week",
    "M": "month",
    "Y": "year",
}


class UnitConversion:
    class Acceleration:
        """Creates an acceleration object that can store an acceleration value and
        convert between units of acceleration."""

        ureg.define("m_s2 = meter / (second ** 2)")
        ureg.define("ft_s2 = 381/1250 * m_s2")
        ureg.define("in_s2 = 127/5000 * m_s2")
        ureg.define("cm_s2 = 1/100 * m_s2")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the acceleration to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Angle:
        """Creates an angle object that can store an angle value and
        convert between units of angle."""

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the angle to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class AngularVelocity:

        ureg.define("deg_s = degree / second")
        ureg.define("rad_s = 10800 / 3.1415926535897932384626433 * deg_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class Area:
        """Creates an area object that can store an area value and
        convert between units of area."""

        ureg.define("ft2 = 1 * sq_ft")
        ureg.define("m2 = 1562500/145161 * ft2")
        ureg.define("cm2 = 625/580644 * ft2")
        ureg.define("in2 = 1/144 * ft2")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the area to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class AreaPerLength:
        """Creates an object representing area per length and allows for unit conversions."""

        ureg.define("m2_m = (meter ** 2) / meter")
        ureg.define("in2_ft = 127/60000 * m2_m")
        ureg.define("mm2_m = 1/1000000 * m2_m")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the area per length to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Unitless:
        """Creates an object representing a unitless quantity."""

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Returns the value unchanged, as it is a unitless quantity."""
            return self.value

    class BrineDensity:
        """Creates an object representing brine density and allows for unit conversions.
        - g/cm3: Grams per cubic centimeter
        - kg/m3: Kilograms per cubic meter
        - lb/ft3: lb per cubic foot
        - lb/in3: lb per cubic inch
        - ppg: lb per gallon
        - mg/L: Milligrams per liter
        """

        ureg.define("g_cm3 = gram / (centimeter ** 3)")
        ureg.define("kg_m3 = 1 / 1000 * g_cm3")
        ureg.define("lb_ft3 = 226796185 / 14158423296 * g_cm3")
        ureg.define("lb_in3 = 226796185 / 8193532 * g_cm3")
        ureg.define("ppg = 2945405 / 24580596 * g_cm3")
        ureg.define("mg_l = 1 / 1000000 * g_cm3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the brine density to the specified unit."""

            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class CGR:
        """Creates and allows CGR unit conversions.
        - CGR: Condensate Gas Ratio
        - scf/scf: Standard cubic feet per standard cubic feet
        - sm3/sm3: Standard cubic meters per standard cubic meter
        - sm3/MMsm3: Standard cubic meters per million standard cubic meters
        - sbbl/MMscf: Stock tank barrels per million standard cubic feet
        - sbbl/scf: Stock tank barrels per standard cubic feet
        """

        ureg.define("scf_scf = ((foot ** 3) / (foot ** 3))")
        ureg.define("sm3_sm3 = 1 * scf_scf")
        ureg.define("sm3_mmsm3 = 1 / 1000000 * scf_scf")
        ureg.define("sbbl_mmscf = 539 / 96000000 * scf_scf")
        ureg.define("sbbl_scf = 539 / 96 * scf_scf")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the CGR to the specified unit."""

            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class CorrosionRate:
        """
        convert between units of corrosion rate.
        - m/s: Meters per second
        - m/yr: Meters per year
        - mm/yr: Millimeters per year
        - mpy: Mil per year (Mil is a unit of length equal to one thousandth of an inch)
        """

        ureg.define("m_s = meter / second")
        ureg.define("m_y = 1 / 31536000 * m_s")
        ureg.define("mm_y = 1 / 31536000000 * m_s")
        ureg.define("mpy = 127 / 157680000000000 * m_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the corrosion rate to the specified unit."""

            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Density:
        """Creates density and allows for unit conversions.
        - g/cm3: Grams per cubic centimeter
        - kg/m3: Kilograms per cubic meter
        - lb/ft3: lb per cubic foot
        - lb/in3: lb per cubic inch
        """

        ureg.define("g_cm3 = 1 * (gram / centimeter ** 3)")
        ureg.define("kg_m3 = 1 / 1000 * g_cm3")
        ureg.define("lb_ft3 = 226796185 / 14158423296 * g_cm3")
        ureg.define("lb_in3 = 226796185 / 8193532 * g_cm3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the density to the specified unit."""

            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class DensityPerPressure:
        """allows density per pressure unit conversions
        - kg/m-N: Kilograms per meter per Newton
        - s2/m2: Seconds squared per square meter
        """

        ureg.define("kg_mn = kilogram / (meter * newton)")
        ureg.define("s2_m2 = second ** 2 / meter ** 2")
        ureg.define("kg_mn = 1 * s2_m2")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the density to the specified unit."""

            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class DensityPerTemperature:
        """allows density per temperature unit conversions
        - kg/m3-C: Kilograms per cubic meter per degree Celsius
        - kg/m3-K: Kilograms per cubic meter per Kelvin
        """

        ureg.define("kg_m3k = kilogram / ((meter ** 3) * kelvin)")
        ureg.define("kg_m3c = 1 * kg_m3k")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the density to the specified unit."""

            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class DiffusionCoefficient:
        """allows DiffusionCoefficient unit conversions
        - cm2/s: square centimeters per second
        - m2/s: square meters per second
        - mm2/s: square millimeters per second
        """

        ureg.define("cm2_s = (centimeter ** 2) / second")
        ureg.define("m2_s = 10000 * cm2_s")
        ureg.define("mm2_s = 1 / 100 * cm2_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the density to the specified unit."""

            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Distance:
        """Creates an angle object that can store an angle value and
        convert between units of angle."""

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the angle to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class DynamicViscosity:
        """Allows Dynamic Viscosity unit conversions.
        - cp: Centipoise
        - kg/m-h: Kilograms per meter per hour
        - kg/m-s: Kilograms per meter per second
        - mPa-s: Millipascal-seconds
        - N-s/m²: Newton-seconds per square meter
        - Pa-s: Pascal-seconds
        - lb/ft-h: lb per foot per hour
        """

        ureg.define("cp = 1 * centipoise")
        ureg.define("pas = 1000 * cp")
        ureg.define("kg_mh = 5/18 * cp")
        ureg.define("kg_ms = 1000 * cp")
        ureg.define("ns_m2 = 1000 * cp")
        ureg.define("mpas = 1 * cp")
        ureg.define("lb_fth = 45359237/109728000 * cp")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the area to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Energy:
        """
        Allows for energy unit conversion:
        - BTU: British Thermal Unit
        - CAL: Calorie
        - J: Joule
        """

        ureg.define("btu = 1 * Btu")
        ureg.define("cal = 0.003965 * btu")
        ureg.define("j = 0.00094823 * btu")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the angle to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Enthalpy:
        """
        Allows for Enthalpy unit conversion:
        - BTU: British Thermal Unit
        - CAL: Calorie
        - J: Joule
        """

        ureg.define("btu = 1 * Btu")
        ureg.define("cal = 0.003965 * btu")
        ureg.define("j = 0.00094823 * btu")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the angle to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class EnthalpyPerMass:
        """
        Allows for EnthalpyPerMass unit conversion:
        British Thermal Units per Pound (btu/lb)
        Joules per Kilogram (J/kg)
        Kilojoules per Kilogram (kJ/kg)
        """

        ureg.define("btu_lb = 1 * (Btu /pound) ")
        ureg.define("j_kg = 4123567 / 9585002404 * btu_lb")
        ureg.define("kj_kg = 1030891750 / 2396250601 * btu_lb")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the angle to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class EnthalpyPerMol:
        """
        Allows for Enthalpy Per Mol unit conversion:
        British Thermal Units per Pound-mole (btu/lbmol)
        Calories per Mole (cal/mol)
        Joules per Mole (J/mol)
        """

        ureg.define("j_mol = joule / mole")
        ureg.define("cal_mol = 523/125 * j_mol")
        ureg.define("btu_lbmol = 2396250601/1030891750 * j_mol")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the enthalpy per mol to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class EnthalpyPerMolPerTemperature:
        """
        Allows for Enthalpy Per Mol Per Temperature unit conversion:
        British Thermal Units per Pound-mole per Rankine (btu/lbmol-R)
        Joules per Mole per Kelvin (J/mol-K)
        """

        ureg.define("btu_lbmolr = 1 * (Btu / (pound * mole * rankine))")
        ureg.define("j_molk = 5154458750 / 21566255409 * btu_lbmolr")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the enthalpy per mol per temperature to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class EnthalpyPerVolume:
        """
        Allows for Enthalpy Per Volume unit conversion:
        British Thermal Units per Cubic Foot (btu/ft^3)
        Calories per Cubic Meter (cal/m^3)
        Joules per Cubic Meter (J/m^3)
        """

        ureg.define("btu_ft3 = 1* (Btu / (foot ** 3))")
        ureg.define("cal_m3 = 231401730744 / 2059277860234375 * btu_ft3")
        ureg.define("j_m3 = 442450728 / 16474222881875 * btu_ft3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the enthalpy per volume to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Entropy:
        """
        Allows for Entropy unit conversion:
        British Thermal Units per Rankine (btu/R)
        Calories per Kelvin (cal/K)
        Joules per Kelvin (J/K)
        """

        ureg.define("btu_r = 1 * (Btu / rankine)")
        ureg.define("cal_K = 523000000 / 237228809499 * btu_r")
        ureg.define("j_K = 125000000 / 237228809499 * btu_r")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the entropy to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class EntropyPerMass:
        """
        Allows for Entropy Per Mass unit conversion:
        British Thermal Units per Pound per Fahrenheit (btu/lb-F)
        Joules per Kilogram per Celsius (J/kg-C)
        Kilojoules per Kilogram per Celsius (kJ/kg-C)
        """

        ureg.define("j_kgc = 1 * (joule / (kilogram * degC))")
        ureg.define("btu_lbf = 86265021636/20617835 * j_kgc")
        ureg.define("kj_kgc = 1000 * j_kgc")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the entropy per mass to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class EntropyPerMol:
        """
        Allows for Entropy Per Mol unit conversion:
        British Thermal Units per Pound per Rankine (btu/lbmol-R)
        Calories per Mol per Kelvin (cal/mol-K)
        Joules per Mol per Kelvin (J/mol-K)
        """

        ureg.define("j_molk = 1 * (joule / (mole * degK))")
        ureg.define("cal_molk = 523/125 * j_molk")
        ureg.define("btu_lbmolr = 21566255409/5154458750 * j_molk")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the entropy per mol to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class EntropyPerVolume:
        """
        Allows for Entropy Per Volume unit conversion:
        British Thermal Units per Cubic Foot per Rankine (btu/ft^3-R)
        Calories per Cubic Meter per Kelvin (cal/m^3-K)
        Joules per Cubic Meter per Kelvin (J/m^3-K)
        """

        ureg.define("j_m3k = joule / ((meter ** 3) * degK)")
        ureg.define("cal_m3k = 523/125 * j_m3k")
        ureg.define("btu_ft3r = 411855572046875/25711303416 * j_m3k")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the entropy per volume to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class ErosionRate:
        """
        Allows for Erosion Rate unit conversion:
        Inches per Year (in/yr)
        Millimeters per Year (mm/yr)
        Inches per Thousand Years (in/kyr)
        """

        ureg.define("in_y = 1 * (inch / year)")
        ureg.define("mm_y = 5 / 127 * in_y")
        ureg.define("in_kyr = 1 / 1000 * in_y")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the erosion rate to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Force:
        """
        Allows for Force unit conversion:
        Dyne (dyn)
        Kilogram-force (kgf)
        Newton (N)
        Pound-force (lbf)
        """

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the angle to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Fraction:
        """
        Allows Fraction unit conversions:
        Parts per billion (ppb)
        Parts per million (ppm)
        Percent (%)
        Dimensionless (-)
        """

        ureg.define("ppm = 1 * ppm")
        ureg.define("ppb = 1 / 1000 * ppm")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert_to(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Frequency:
        """
        Allows Frequency unit conversions:
        Hertz (Hz)
        Inverse hour (1/h)
        Inverse minute (1/min)
        Inverse second (1/s)
        """

        ureg.define("hz = 1 * hertz")
        ureg.define("one_h = 1 / 3600 * hz")
        ureg.define("one_m = 1 / 60 * hz")
        ureg.define("one_s = 1 * hz")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class FrictionFactor:
        """
        Allows Friction Factor unit conversions:
        Newton-second per meter (N-s/m)
        """

        ureg.define("ns_m = (newton * second) / meter")
        ureg.define("g_s = 1 / 1000 * ns_m")
        ureg.define("kg_h = 1 / 3600 * ns_m")
        ureg.define("kg_s = 1 * ns_m")
        ureg.define("kt_d = 625 / 54 * ns_m")
        ureg.define("lb_h = 45359237 / 360000000000 * ns_m")
        ureg.define("lb_s = 45359237 / 100000000 * ns_m")
        ureg.define("t_d = 5 / 432 * ns_m")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class GasFlowrate:
        """
        Allows Gas Flowrate unit conversions:
        Barrel per day (bbl/d)
        Barrel per hour (bbl/h)
        Barrel per minute (bbl/min)
        Cubic foot per day (ft3/d)
        Cubic foot per hour (ft3/h)
        Cubic foot per minute (ft3/min)
        Cubic foot per second (ft3/s)
        Cubic meter per day (m3/d)
        Cubic meter per hour (m3/h)
        Cubic meter per second (m3/s)
        Gallon per minute (gpm)
        Liter per minute (L/min)
        Milliliter per minute (mL/min)
        """

        ureg.define("m3_s = (meter ** 3) / second")
        ureg.define("bbl_d = 1104078437 / 600000000000000 * m3_s")
        ureg.define("bbl_h = 1104078437 / 25000000000000 * m3_s")
        ureg.define("bbl_min = 3312235311 / 1250000000000 * m3_s")
        ureg.define("ft3_d = 2048383 / 6250000000000 * m3_s")
        ureg.define("ft3_h = 6145149 / 781250000000 * m3_s")
        ureg.define("ft3_min = 18435447 / 39062500000 * m3_s")
        ureg.define("ft3_s = 55306341 / 1953125000 * m3_s")
        ureg.define("m3_d = 1 / 86400 * m3_s")
        ureg.define("m3_h = 1 / 3600 * m3_s")
        ureg.define("gpm = 157725491 / 2500000000000 * m3_s")
        ureg.define("l_min = 1 / 60000 * m3_s")
        ureg.define("ml_min = 1 / 60000000 * m3_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class GasStandardFlowrate:
        """
        Allows Gas Standard Flowrate unit conversions:
        Million standard cubic feet per day (MMscf/d)
        Standard cubic feet per day (scf/d)
        Standard cubic feet per hour (scf/h)
        Standard cubic feet per second (scf/s)
        Standard cubic meter per day (sm3/d)
        Standard cubic meter per hour (sm3/h)
        Standard cubic meter per second (sm3/s)
        Thousand standard cubic feet per day (Mscf/d)
        Thousand standard cubic meters per day (Msm3/d)
        Thousand standard cubic meters per hour (Msm3/h)
        """

        ureg.define("sm3_s = (meter ** 3) / second")
        ureg.define("mmscf_d = 2048383/6250000 * sm3_s")
        ureg.define("scf_d = 2048383/6250000000000 * sm3_s")
        ureg.define("scf_h = 6145149/781250000000 * sm3_s")
        ureg.define("scf_s = 55306341/1953125000 * sm3_s")
        ureg.define("sm3_d = 1/86400 * sm3_s")
        ureg.define("sm3_h = 1/3600 * sm3_s")
        ureg.define("mscf_d = 2048383/6250000000 * sm3_s")
        ureg.define("msm3_d = 625/54 * sm3_s")
        ureg.define("msm3_h = 2500/9 * sm3_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class GLR:
        """
        Allows Gas-Liquid Ratio unit conversions:
        MMscf/sbbl (Million standard cubic feet per stock barrel)
        scf/scf (Standard cubic feet per standard cubic feet)
        scf/sbbl (Standard cubic feet per stock barrel)
        sm3/sm3 (Standard cubic meter per standard cubic meter)
        """

        ureg.define("sm3_sm3 = ((meter ** 3) / (meter ** 3))")
        ureg.define("mmscf_sbbl = 96000000 / 539 * sm3_sm3")
        ureg.define("scf_scf = 1 * sm3_sm3")
        ureg.define("scf_sbbl = 96 / 539 * sm3_sm3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class GOR:
        """
        Allows Gas-Oil Ratio unit conversions:
        MMscf/sbbl (Million standard cubic feet per stock barrel)
        scf/scf (Standard cubic feet per standard cubic feet)
        scf/sbbl (Standard cubic feet per stock barrel)
        sm3/sm3 (Standard cubic meter per standard cubic meter)
        """

        ureg.define("sm3_sm3 = ((meter ** 3) / (meter ** 3))")
        ureg.define("mmscf_sbbl = 96000000 / 539 * sm3_sm3")
        ureg.define("scf_scf = 1 * sm3_sm3")
        ureg.define("scf_sbbl = 96 / 539 * sm3_sm3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class HeatFlux:
        """
        Allows Heat Flux unit conversions:
        btu/ft2-s (British thermal unit per square foot per second)
        btu/in2 (British thermal unit per square inch per second)
        kW/m2 (Kilowatt per square meter)
        W/cm2 (Watt per square centimeter)
        W/m2 (Watt per square meter)
        """

        ureg.define("w_m2 = watt / (meter ** 2)")
        ureg.define("btu_ft2s = 26358756611/2322576 * w_m2")
        ureg.define("btu_in2s = 26358756611/16129 * w_m2")
        ureg.define("kw_m2 = 1000 * w_m2")
        ureg.define("w_cm2 = 10000 * w_m2")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class HeatTransferCoefficient:
        """
        Allows Heat Transfer Coefficient unit conversions:
        btu/ft2-h-F (British thermal unit per square foot per hour per Fahrenheit degree)
        cal/m2-h-C (Calorie per square meter per hour per Celsius degree)
        W/m2-C (Watt per square meter per Celsius degree)
        W/m2-K (Watt per square meter per Kelvin degree)
        """

        ureg.define("w_m2c = watt / (meter ** 2) * degC")
        ureg.define("W_m2k = 1 * w_m2c")
        ureg.define("btu_ft2hf = 26358756611/4645152000 * w_m2c")
        ureg.define("cal_m2hc = 523/450000 * w_m2c")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class HeatTransferPerLength:
        """
        Allows Heat Transfer Per Length unit conversions:
        British Thermal Unit per second per foot (btu/s-ft)
        Kilowatt per meter (kW/m)
        Watt per meter (W/m)
        """

        ureg.define("w_m = 1 * (watt / meter)")
        ureg.define("btu_sft = 26358756611 / 7620000 * w_m")
        ureg.define("kw_m = 1000 * w_m")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class KinematicViscosity:
        """
        Allows Kinematic Viscosity unit conversions:
        Centistokes (cSt)
        Square foot per hour (ft2/h)
        Square inch per second (in2)
        Square meter per second (m2/s)
        """

        ureg.define("m2_s = (meter ** 2) / second")
        ureg.define("cst =  1 / 1000000 * m2_s")
        ureg.define("ft2_h = 16129 / 625000000 *m2_s")
        ureg.define("in2_s = 16129 / 25000000 * m2_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Length:
        """
        Allows Length unit conversions:
        Centimeter (cm)
        Foot (ft)
        Inch (in)
        Kilometer (km)
        Meter (m)
        Micrometer (µm)
        Mile (mi)
        Millimeter (mm)
        Yard (yd)
        """

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class LinearMomentum:
        """
        Allows Linear Momentum unit conversions:
        Kilogram-meter per second (kg-m/s)
        Newton-second (N-s)
        """

        ureg.define("ns = 1 * (newton * second)")
        ureg.define("kgm_s = 1 * ns")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class LiquidFlowRate:
        """
        Allows Liquid Flowrate unit conversions:
        Barrels per day (bbl/d)
        Barrels per hour (bbl/h)
        Barrels per minute (bbl/min)
        Cubic feet per day (ft3/d)
        Cubic feet per hour (ft3/h)
        Cubic feet per minute (ft3/min)
        Cubic feet per second (ft3/s)
        Cubic meters per day (m3/d)
        Cubic meters per hour (m3/h)
        Cubic meters per second (m3/s)
        Gallons per minute (gpm)
        Liters per minute (L/min)
        Milliliters per minute (mL/min)
        """

        ureg.define("m3_s = (meter ** 3) / second")
        ureg.define("bbl_d = 1104078437 / 600000000000000 * m3_s")
        ureg.define("bbl_h = 1104078437 / 25000000000000 * m3_s")
        ureg.define("bbl_min = 3312235311 / 1250000000000 * m3_s")
        ureg.define("ft3_d = 2048383 / 6250000000000 * m3_s")
        ureg.define("ft3_h = 6145149 / 781250000000 * m3_s")
        ureg.define("ft3_min = 18435447 / 39062500000 * m3_s")
        ureg.define("ft3_s = 55306341 / 1953125000 * m3_s")
        ureg.define("m3_d = 1 / 86400 * m3_s")
        ureg.define("m3_h = 1 / 3600 * m3_s")
        ureg.define("gpm = 157725491 / 2500000000000 * m3_s")
        ureg.define("l_min = 1 / 60000 * m3_s")
        ureg.define("ml_min = 1 / 60000000 * m3_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class LiquidStandardFlowRate:
        """
        Allows Liquid Standard Flowrate unit conversions:
        Standard cubic meters per day (sm3/d)
        Standard cubic meters per hour (sm3/h)
        Standard cubic meters per second (sm3/s)
        Standard barrels per day (sbbl/d)
        Standard barrels per hour (sbbl/h)
        Standard barrels per minute (sbbl/min)
        """

        ureg.define("sm3_s = (meter ** 3) / second")
        ureg.define("sm3_d = 1 / 86400 * sm3_s")
        ureg.define("sm3_h = 1 / 3600 * sm3_s")
        ureg.define("sbbl_d = 1104078437 / 600000000000000* sm3_s")
        ureg.define("sbbl_h = 1104078437 / 25000000000000 * sm3_s")
        ureg.define("sbbl_min = 3312235311 / 1250000000000 * sm3_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Mass:
        """
        Allows Mass unit conversions:
        Gram (g)
        Kilogram (kg)
        Milligram (mg)
        Kiloton (kt)
        Kip (kips)
        Ounce (oz)
        Pound (lb)
        Ton (t)
        """

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the angle to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MassFlowrate:
        """
        Allows Mass Flowrate unit conversions:
        Gram per second (g/s)
        Kilogram per hour (kg/h)
        Kilogram per second (kg/s)
        Kiloton per day (kt/d)
        Pound per hour (lb/h)
        Pound per second (lb/s)
        Ton per day (t/d)
        """

        ureg.define("g_s = gram / second")
        ureg.define("kg_h = 5/18 * g_s")
        ureg.define("kg_s = 1000 * g_s")
        ureg.define("kt_d = 312500/27 * g_s")
        ureg.define("lb_h = 45359237/360000000 * g_s")
        ureg.define("lb_s = 45359237/100000 * g_s")
        ureg.define("t_d =  625/54 * g_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the mass flowrate to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MassFlowratePerArea:
        """
        Allows Mass Flowrate Per Area unit conversions:
        Kilogram per hour per square meter (kg/h-m2)
        Kilogram per hour per square centimeter (kg/h-cm2)
        Kilogram per second per square meter (kg/s-m2)
        Pound per hour per square meter (lb/h-m2)
        Pound per second per square meter (lb/s-m2)
        Milligram per hour per square inch (mg/h-in2"""

        ureg.define("kg_hm2 = 1 * (kilogram / hour * (meter ** 2))")
        ureg.define("kg_hcm2 = 10000 * kg_hm2")
        ureg.define("kg_sm2 = 3600 * kg_hm2 ")
        ureg.define("lb_hm2 = 45359237 / 100000000 * kg_hm2")
        ureg.define("lb_sm2 = 408233133 / 250000 * kg_hm2")
        ureg.define("mg_hin2 = 25 / 16129 * kg_hm2")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the mass flowrate per area to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MassGradient:
        """
        Allows Mass Gradient unit conversions:
        Gram per centimeter (g/cm)
        Gram per foot (g/ft)
        Gram per inch (g/in)
        Gram per meter (g/m)
        Kilogram per meter (kg/m)
        Pound per foot (lb/ft)
        Pound per inch (lb/in)
        """

        ureg.define("g_m = 1 * (gram / meter)")
        ureg.define("g_cm = 100 * g_m")
        ureg.define("g_ft = 1250 / 381 * g_m")
        ureg.define("g_in = 5000 / 127 * g_m")
        ureg.define("kg_m = 1000 * g_m")
        ureg.define("lb_ft = 45359237 / 30480 * g_m")
        ureg.define("lb_in = 45359237 / 2540 * g_m")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MassMomentOfInertia:
        """
        Allows Mass Moment of Inertia unit conversions:
        Kilogram square meter (kg-m^2)
        Pound foot square (lb-ft^2)
        Pound force inch second square (lbf-in-s^2)
        Pound inch square (lb-in^2)
        """

        ureg.define("kgm2 = kilogram * (meter ** 2)")
        ureg.define("lbft2 = 6584392202157/156250000000000 * kgm2")
        ureg.define("lbfins2 = 1129848290276167/10000000000000000 * kgm2")
        ureg.define("lbin2 = 731599133573/2500000000000000 * kgm2")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MassPerArea:
        """
        Allows Mass Per Area unit conversions:
        Gram per square centimeter (g/cm^2)
        Gram per square foot (g/ft^2)
        Gram per square inch (g/in^2)
        Gram per square meter (g/m^2)
        Kilogram per square foot (kg/ft^2)
        Kilogram per square meter (kg/m^2)
        """

        ureg.define("g_m2 = gram / (meter ** 2)")
        ureg.define("g_cm2 = 10000 * g_m2")
        ureg.define("g_ft2 = 1562500 / 145161 * g_m2")
        ureg.define("g_in2 = 25000000 / 16129 * g_m2")
        ureg.define("kg_ft2 = 1562500000 / 145161 * g_m2")
        ureg.define("kg_m2  = 1000 * g_m2")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MolarConcentrationGradient:
        """
        Allows Molar Concentration Gradient unit conversions:
        Mole per cubic centimeter to the power of 4 (mol/cm^4)
        Mole per cubic meter to the power of 4 (mol/m^4)
        Pound-mole per cubic foot to the power of 4 (lbmol/ft^4)
        """

        ureg.define("mol_m4 = mole / (meter ** 4)")
        ureg.define("mol_cm4 = 100000000 * mol_m4")
        ureg.define("lbmol_ft4 = 2214806494140625 / 42143431842 * mol_m4")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MolarDensity:
        """
        Allows Molar Density unit conversions:
        Mole per cubic centimeter (mol/cm^3)
        Mole per cubic meter (mol/m^3)
        Mole per liter (mol/L)
        Pound-mole per cubic foot (lbmol/ft^3)
        """

        ureg.define("mol_m3 = mole / (meter ** 3)")
        ureg.define("mol_l = 1000 * mol_m3")
        ureg.define("lbmol_ft3 = 3543690390625/221225364 * mol_m3")
        ureg.define("mol_cm3 = 1000000 * mol_m3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class MolarFlux:
        """
        Allows Molar Flux unit conversions:
        Kilomole per square meter per second (kmol/m^2-s)
        Mole per square meter per second (mol/m^2-s)
        """

        ureg.define("mol_m2s = 1 * (mole / ((meter ** 2) * second))")
        ureg.define("kmol_m2s = 1000 * mol_m2s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            """Converts the value to the specified unit."""
            converted_quantity = self.value * ureg(self.unit).to(to_unit)
            return converted_quantity.magnitude

    class Volume:
        ureg.define("m3 = meter ** 3")
        ureg.define("ft3 = 55306341/1953125000 * m3")
        ureg.define("gal = 473176473/125000000000 * m3")
        ureg.define("l = 1/1000 * m3")
        ureg.define("mmft3 = 442450728/15625 * m3")
        ureg.define("bbl = 9936705933/62500000000 * m3")
        ureg.define("cm3 = 1/1000000 * m3")
        ureg.define("ml = 1/1000000 * m3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class VolumetricFlowRate:
        ureg.define("m3_s =  (meter ** 3) / second")
        ureg.define("gpm = 157725491/2500000000000 * m3_s")
        ureg.define("m3_h = 3600 * m3_s")
        ureg.define("m3_d =  86400 * m3_s")
        ureg.define("ft3_s = 55306341/1953125000 * m3_s")
        ureg.define("ft3_m= 18435447/39062500000 * m3_s")
        ureg.define("ft3_h= 6145149/781250000000 * m3_s ")
        ureg.define("ft3_d = 2048383/6250000000000 * m3_s")
        ureg.define("bbl_d = 1104078437/600000000000000 * m3_s")
        ureg.define("bbl_h = 1104078437/25000000000000 * m3_s")
        ureg.define("bbl_m = 3312235311/1250000000000 * m3_s")
        ureg.define("l_m =  1/60000 * m3_s")
        ureg.define("ml_m = 1/60000000 * m3_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class VolumetricStandardFlowrate:
        """Creates an object representing standard volumetric flowrate and allows for unit conversions.
        - MMscf/d: Million standard cubic feet per day
        - scf/d: Standard cubic feet per day
        - scf/h: Standard cubic feet per hour
        - scf/s: Standard cubic feet per second
        - sm3/d: Standard cubic meters per day
        - sm3/h: Standard cubic meters per hour
        - sm3/s: Standard cubic meters per second
        - sbbl/d: Stock tank barrels per day
        - sbbl/h: Stock tank barrels per hour
        - sbbl/min: Stock tank barrels per minute
        - Mscf/d: Thousand standard cubic feet per day
        - Msm3/d: Thousand standard cubic meters per day
        - Msm3/h: Thousand standard cubic meters per hour
        """

        ureg.define("sm3_s = (meter ** 3) / second")
        ureg.define("mmscf_d = 2048383 / 6250000 * sm3_s")
        ureg.define("scf_d = 2048383 / 6250000000000  * sm3_s")
        ureg.define("scf_h = 6145149 / 781250000000 * sm3_s")
        ureg.define("scf_s = 55306341 / 1953125000 * sm3_s")
        ureg.define("sm3_d = 1 / 86400 * sm3_s")
        ureg.define("sm3_h = 1 / 3600 * sm3_s")
        ureg.define("mscf_d = 2048383 / 6250000000 * sm3_s")
        ureg.define("msm3_d = 625 / 54 * sm3_s")
        ureg.define("msm3_h = 2500 / 9 * sm3_s")
        ureg.define("sbbl_h = 1104078437 / 25000000000000 * sm3_s")
        ureg.define("sbbl_min = 3312235311 / 1250000000000 * sm3_s")
        ureg.define("sbbl_d = 1104078437 / 600000000000000 * sm3_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class StandardVolume:
        """Creates an object representing standard volume and allows for unit conversions.
        scm3 - Standard Cubic Meter
        MMscf - Million Standard Cubic Feet
        sgal - Standard Gallon
        sL - Standard Liter
        scf - Standard Cubic Feet
        sm3 - Standard Cubic Meter
        sbbl - Standard Barrel
        Mscf - Thousand Standard Cubic Feet
        Msm3 - Thousand Standard Cubic Meters
        """

        ureg.define("sm3 = 1 * (meter ** 3)")
        ureg.define("mmscf =   442450728 / 15625  * sm3")
        ureg.define("scf = 55306341 / 1953125000 * sm3")
        ureg.define("scm3 = 1 / 1000000 * sm3")
        ureg.define("mscf = 55306341 / 1953125 * sm3")
        ureg.define("sl = 1 / 1000  * sm3")
        ureg.define("sgal = 473176473 / 125000000000 * sm3")
        ureg.define("sbbl = 9936705933 / 62500000000 * sm3")
        ureg.define("msm3 = 1000000 * sm3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class StandardVolumetricFlowratePerLength:
        """Creates an object representing standard volumetric flowrate per length and allows for unit conversions."""

        ureg.define("m2_s = 1 * ((meter ** 2) / second)")
        ureg.define("sbbl_d_ft = 8693531 / 1440000000000 * m2_s")
        ureg.define("mm2_s = 1 / 1000000 * m2_s")
        ureg.define("cm2_s = 1 / 10000 * m2_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class Velocity:
        ureg.define("m_s = 1 * (meter / second)")
        ureg.define("m_h = 1/3600 * m_s")
        ureg.define("mph = 1397/3125 * m_s")
        ureg.define("km_h = 5/18 * m_s")
        ureg.define("km_s = 1000 * m_s")
        ureg.define("in_h = 127/18000000 * m_s")
        ureg.define("in_s = 127/5000 * m_s")
        ureg.define("ft_h = 127/1500000 * m_s")
        ureg.define("ft_s = 381/1250 * m_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class Torque:
        ureg.define("ftlb = 1 * (foot * pound)")
        ureg.define("mnm = 3389544870828501/2500000000000 * ftlb")
        ureg.define("ncm = 3389544870828501/25000000000000 * ftlb")
        ureg.define("nm = 3389544870828501/2500000000000000 * ftlb")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class Time:
        ureg.define("ms = 1/1000 * second")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class ThermalExpansion:
        ureg.define("one_c = 1/ degC")
        ureg.define("one_k = 1/ degK")
        ureg.define("one_f = 1/ degF")
        ureg.define("one_r = 1/ degR")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class ThermalConductivity:

        ureg.define("w_mc = watt / (meter * celsius)")
        ureg.define("btu_fthr = 26358756611 / 15240000000 * w_mc")
        ureg.define("cal_mhk = 523 / 450000 * w_mc")
        ureg.define("w_mk = 1 * w_mc")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class TemperatureGradient:
        ureg.define("f_ft = degF/ foot")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class TemperatureDifference:

        ureg.define("dtc = 1 * delta_degC")
        ureg.define("dtk = 1 * dtc")
        ureg.define("dtf = 5/9 * dtc")
        ureg.define("dtr = 5/9 * dtc")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            unit_map = {"c": "dtc"}
            if self.unit in unit_map:
                self.unit = unit_map[self.unit]
                converted = self.value * ureg(self.unit).to(to_unit)
                return converted.magnitude

    class Temperature:
        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            unit_map = {
                "c": "celsius",
            }
            if self.unit in unit_map:
                self.unit = unit_map[self.unit]
                temp = Q_(self.value, self.unit)
                converted = temp.to(to_unit)
                return converted.magnitude

    class SurfaceTension:
        ureg.define("n_m = N / meter")
        ureg.define("dyn_cm = 1/1000 * n_m")
        ureg.define("mn_m = 1/1000 * n_m")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class SpecificVolume:
        ureg.define("m3_kg = ((meter ** 3) / kilogram)")
        ureg.define("ft3_lb = 1769802912/28349523125 * m3_kg")
        ureg.define("in3_lb = 2048383/56699046250 * m3_kg")
        ureg.define("cm3_g = 1/1000 * m3_kg")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class SpecificMole:
        ureg.define("mol_g = mole / gram")
        ureg.define("mol_kg = 1/1000 * mol_g")
        ureg.define("lbmol_lb = 1 * mol_g")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class SpecificHeatCapacity:
        ureg.define("j_kgc = joule / (kilogram * degC)")
        ureg.define("btu_lbf = 86265021636/20617835 * j_kgc")
        ureg.define("btu_lbr = 86265021636/20617835 * j_kgc")
        ureg.define("j_kgk = 1 * j_kgc")
        ureg.define("kj_kgc = 1000 * j_kgc")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class SpecificEnthalpy:
        ureg.define("j_g = joule / gram")
        ureg.define("j_kg = 1/1000 * j_g")
        ureg.define("kj_kg = 1 * j_g")
        ureg.define("btu_lb = 2396250601/1030891750 * j_g")
        ureg.define("cal_kg = 523/125000 * j_g")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class SpecificCapacity:
        ureg.define("m3_r = (meter **3) / degR")
        ureg.define("ft3_r = 55306341/1953125000 * m3_r")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class RevolutionsPerTime:
        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class RateOfChangeOfTemperature:
        ureg.define("c_s = celsius / second")
        ureg.define("f_s = 5/9 * c_s")
        ureg.define("k_s = 1 * c_s")
        ureg.define("r_s = 5/9 * c_s")
        ureg.define("c_m = 1/60 * c_s")
        ureg.define("f_m = 1/108 * c_s")
        ureg.define("c_h = 1/3600 * c_s")
        ureg.define("f_h = 1/6480 * c_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class RateOfChangeOfPressure:
        ureg.define("pa_s = Pa / second")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class RateOfChangeOfDensity:
        ureg.define("kg_m3s = kilogram / (meter ** 3) * second")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class PressureGradient:
        ureg.define("pa_m = Pa / meter")
        ureg.define("psi_ft = 8896443230521/393289536 * pa_m")
        ureg.define("kpa_m = 1000 * pa_m")
        ureg.define("bar_m = 1 * 100000 * pa_m")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class PressureDifference:
        ureg.define("pa = 1 * Pa")
        ureg.define("kpa = 1000 * pa")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class Pressure:
        ureg.define("pa = 1 * Pa")
        ureg.define("kpa = 1000 * pa")
        ureg.define("barg = 100000 * pa")
        ureg.define("bara = 100000 * pa")
        ureg.define("bar = 100000 * pa")
        ureg.define("atm = 101325 * pa ")
        ureg.define("mpa = 1000000 * pa")
        ureg.define("psi = 8896443230521 / 1290320000 * pa")
        ureg.define("psia = 8896443230521 / 1290320000 * pa")
        ureg.define("psig = 8896443230521 / 1290320000 * pa")
        ureg.define("kgf_cm2 = 196133 / 2 * pa")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            if to_unit == "psig":
                convert = self.value * ureg(self.unit).to(ureg.psia)
                converted = convert.magnitude - 14.7
                return converted
            elif to_unit == "barg":
                convert = self.value * ureg(self.unit).to(ureg.bara)
                converted = convert.magnitude - 1.01
                return converted
            else:
                converted = self.value * ureg(self.unit).to(to_unit)
                return converted.magnitude

    class Power:
        ureg.define("w = 1 * watt")
        ureg.define("btu_h = 26358756611/90000000000 * w")
        ureg.define("hp = 37284993579113511/50000000000000 * w")
        ureg.define("kw = 1000 * w")
        ureg.define("mw = 1/1000 * w")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class OneOverLength:
        ureg.define("one_m = 1 / meter")
        ureg.define("one_cm = 100 * one_m")
        ureg.define("one_ft = 1250/381 * one_m")
        ureg.define("one_km = 1/1000 * one_m")
        ureg.define("one_mi = 125/201168 * one_m")
        ureg.define("one_mm = 1000 * one_m")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class OneOverFlowrate:
        ureg.define("d_m3 = d / (meter ** 3)")
        ureg.define("d_bbl = 62500000000/9936705933 * d_m3")
        ureg.define("min_bbl = 390625000/89430353397 * d_m3")
        ureg.define("d_ft3 = 1953125000/55306341 * d_m3")
        ureg.define("h_ft3 = 244140625/165919023 * d_m3")
        ureg.define("s_ft3 = 9765625/23892339312 * d_m3")
        ureg.define("h_m3 = 1/24* d_m3")
        ureg.define("s_m3 = 1/86400 * d_m3")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class Moles:
        ureg.define("mol = 1 * mole")
        ureg.define("kmol = 1000 * mol")
        ureg.define("lbmol = 45359237/100000 * mol")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class MoleDensityPerTemperature:
        ureg.define("mol_m3k = mole / ((meter ** 3) * degK)")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class MolecularWeight:
        ureg.define("g_mol = gram / mole")
        ureg.define("kg_kmol = 1 * g_mol")
        ureg.define("kg_mol = 1000 * g_mol")
        ureg.define("lb_lbmol = 1 * g_mol")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class MolarVolume:
        ureg.define("m3_mol = ((meter ** 3) / mole)")
        ureg.define("cm3_mol = 1/1000000 * m3_mol")
        ureg.define("l_mol = 1/1000 * m3_mol")
        ureg.define("ft3_lbmol = 221225364/3543690390625 * m3_mol")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude

    class MolarRate:
        ureg.define("mol_s = mole / second")
        ureg.define("kmol_s = 1000 * mol_s")
        ureg.define("lbmol_s = 45359237/100000 * mol_s")

        def __init__(self, value, unit):
            self.value = value
            self.unit = unit

        def convert(self, to_unit):
            converted = self.value * ureg(self.unit).to(to_unit)
            return converted.magnitude



"""
TPL parser functionality
"""

import warnings

# temporary to remove pandas 3.0 deprecationwarning about pyarrow from printing
warnings.filterwarnings("ignore", category=DeprecationWarning)

import pandas as pd
import argparse
import re
from functools import reduce
from pathlib import Path
import os
import sys
import math
import yaml


sys.path.insert(0, "src")
from pypontem.utils.unit_conversion import UnitConversion, unit_map

current_dir = os.path.dirname(os.path.abspath(__file__))
YAML_PATH = os.path.join(current_dir, "..", "utils", "units.yaml")

# YAML_PATH = "src/pypontem/utils/units.yaml"


def read_file(file_path):
    if os.path.isfile(file_path):
        with open(file_path, "r") as file:
            content = file.read()
        return content
    else:
        raise FileNotFoundError(
            f"The file specified '{file_path}' does not exist. Please verify the location of the file."
        )


def open_yaml(yaml_path):
    with open(yaml_path, "r") as file:
        unitsdb = yaml.safe_load(file)
    return unitsdb


pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
pd.set_option("display.max_colwidth", None)


def search(df, var_name=None, loc_name=None, pipe_name=None):
    r"""
    Searches for variables containing a keyword in their names within a DataFrame.
        Args:
            df (pandas.DataFrame): The DataFrame to search within.
            keyword (str): The keyword to search for.

        Returns:
            None
    """
    filter_conditions = []
    if var_name:
        var_name = var_name.upper()
        filter_conditions.append(df["varname"] == var_name)
    if loc_name:
        loc_name = loc_name.upper()
        filter_conditions.append(df["locname"] == loc_name)
    if pipe_name:
        pipe_name = pipe_name.upper()
        filter_conditions.append(
            df["pipename"].str.contains(pipe_name, na=False, case=False)
        )
    if len(filter_conditions) > 1:
        result_df = df[reduce(lambda x, y: x & y, filter_conditions)]
    else:
        result_df = df[filter_conditions[0]]
    return result_df


class tplParser:
    """
    Class which holds the tpl parser functions

    Functions:
        1. metadata: to extract metadata from a tpl file
        2. branch_names: to extract branch names from a tpl file
        3. branch_profiles: to extract branch profile information from a tpl file
        4. parse_number_of_variables: to extract the number of variables present in a tpl file
        5. extract_catalog: to extract the catalog information from a tpl file
        6. search_catalog: to extract the information from the catalog of the variable specified
        7. extract_trends: to extract trends in a tpl file
        8. calc_average: to compute the average of trends extracted in a tpl file	
    """


    def __init__(self, filepath):
        """
        Initializes an instance of the class by loading file data and unit definitions.
        Arguments:
            - filepath (str): The path to the file to be loaded.
        Returns:
            - None
        """

        # read file from path
        self.filepath = Path(filepath)
        # save content of file in memory so we don't open it repeatedly - NOTE: may need to be changed later
        self.content = read_file(self.filepath)
        # open unit yaml file
        self.unitsdb = open_yaml(YAML_PATH)

    @property
    def metadata(self): 
        """
        Extracts metadata from the content stored in the object and returns it as a pandas DataFrame.
        
        Parameters:
        - None
        
        Returns:
            - pandas.DataFrame: A DataFrame containing the extracted metadata with keys as column headers
        """

        return self._extract_metadata()

    def _extract_metadata(self):
        metadata = {}  # Dictionary to store metadata

        try:
            branch_index = self.content.index("BRANCH\n")
        except ValueError:
            branch_index = len(self.content)

        # Extract metadata before "BRANCH"
        metadata_lines = self.content[:branch_index]

        # Parsing specific metadata and storing in dictionary
        for i, line in enumerate(metadata_lines.splitlines()):
            line = line.strip()
            olga_match = re.search(r"OLGA\s+(\S+)", line)
            if olga_match:
                metadata["version"] = line.strip("'")
            elif "INPUT FILE" in line:
                metadata["Input file"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "PVT FILE" in line:
                metadata["pvt"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "DATE" in line:
                metadata["time"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "PROJECT" in line:
                metadata["project"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "TITLE" in line:
                metadata["title"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "AUTHOR" in line:
                metadata["author"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "GEOMETRY" in line:
                match = re.search(r"\((.*?)\)", line)
                if match:
                    metadata["geometry"] = [match.group(1)]
        return pd.DataFrame(metadata)

    @property
    def branch_names(self):
        """
        Extracts branch names from the content stored in the object.
        
        Arguments:
            - None.
        
        Returns:
            - list: A list of extracted branch names as strings.
        """

        return self._extract_branch_names()

    def _extract_branch_names(self):
        df_branch_names = []
        pattern = "(BRANCH|ANNULUS)\n(.+)"
        branch_names = re.findall(pattern, self.content)
        for branch in branch_names:
            df_branch_names.append(
                branch[1]
            )  # Accessing the second element (branch name) of each tuple
        return df_branch_names

    @property
    def branch_profiles(self):
        """
        Extracts and displays elevation for a specific branch or all branches in the file.
        
        Arguments:
            - None.
        
        Returns:
            - Pandas.Dataframe: a dataframe containing the profiles of the specified branch.
        """

        return self._extract_branch_profiles(target_branch=None)

    def _extract_branch_profiles(self, target_branch=None):
        def create_branch_table(branch_data):
            # Ensure lengths and elevations have the same length
            min_length = min(
                len(branch_data["Lengths"]), len(branch_data["Elevations"])
            )
            df = pd.DataFrame(
                {
                    "Lengths_("
                    + str(self.metadata["geometry"].values[0]).lower()
                    + ")": branch_data["Lengths"][:min_length],
                    "Elevations_("
                    + str(self.metadata["geometry"].values[0]).lower()
                    + ")": branch_data["Elevations"][:min_length],
                }
            )
            return df

        # Define a pattern to match each 'BRANCH' section
        branch_pattern = re.compile(
            r"(BRANCH|ANNULUS)\n'(.*?)'\n(\d+)\n([\s\S]*?)(?=(BRANCH|ANNULUS)|\nCATALOG)",
            re.DOTALL,
        )

        # Find all matches
        matches = branch_pattern.findall(self.content)
        found = False  # Flag to check if the target branch is found

        branches = {}

        for match in matches:
            branch_data = {}
            branch_name = match[1]
            num_pipes = int(match[2])
            data_section = match[3].split()

            if data_section:
                midpoint = len(data_section) // 2
                lengths = [float(data) for data in data_section[:midpoint]]
                elevations = [float(data) for data in data_section[midpoint:]]
                branch_data = {
                    "Branch": branch_name,
                    "Lengths": lengths,
                    "Elevations": elevations,
                }

                # Display branch data only if it matches the target branch
                if not target_branch or branch_name == target_branch:
                    found = True  # Set found to True if the branch is found
                    profile = create_branch_table(branch_data)
                    branches[branch_name] = profile

        if (
            not found and target_branch
        ):  # Display message if branch not found and target_branch is provided
            raise ValueError(
                f"The branch '{target_branch}' does not exist in the file."
            )
        else:
            return branches

    @property
    def n_vars(self):
        """
        Parses the number of variables from a given file path.
        
        Arguments:
            - None
        
        Returns:
            - n_vars (int): The number of variables if found, else None.
        """

        return self._parse_number_of_variables()

    def _parse_number_of_variables(self):
        # regex pattern which finds the next line after CATALOG
        pattern = re.search(r"CATALOG[\r \n]+([^\r\n]+)", self.content)
        return int(pattern[1])

    @property
    def catalog(self):
        """
        Extract variable information from the provided content using a regular expression pattern.
        
        Arguments:
            - None.

        Returns:
            - pandas.DataFrame: A DataFrame containing the extracted variable information.
        """

        return self._extract_catalog()

    def _extract_catalog(self):
        pattern = re.compile(
            r"""
            (?P<name>\S+)                      # variable name
            \s+'(?P<locator_type>[^']*)'       # locator type
            (\s+'(?P<locator_name>[^']*)')?    # locator name (optional)
            (\s+'(?P<branch>[^']*)')?          # branch (optional)
            (\s+'(?P<pipe>[^']*)')?            # pipe (optional)
            (\s+'(?P<pipe_name>[^']*)')?       # pipe name (optional)
            (\s+'(?P<pipe_nr>[^']*)')?         # pipe nr (optional)
            (\s+'(?P<pipe_number>[^']*)')?     # pipe number (optional)
            \s+'(?P<unit>[^']*)'               # unit
            \s+'(?P<description>[^']*)'        # description
            """,
            re.IGNORECASE | re.VERBOSE,
        )
        lines = self.content.splitlines()
        catalog_index = next(
            (
                i + 1
                for i, line in enumerate(lines)
                if line.strip() == "CATALOG"
                and i + 1 < len(lines)
                and lines[i + 1].strip().isdigit()
            ),
            None,
        )
        # Extract variable information from the lines after the "CATALOG" line
        matches = pattern.finditer("\n".join(lines[catalog_index:]))
        data_list = []
        # Iterate over matches and extract data
        for match in matches:
            locator_name = match.group("locator_name") or None
            if locator_name == "BRANCH:":
                locator_name = match.group("branch") or None

            out_unit = match.group("unit") or None
            if out_unit:
                out_unit = re.sub(
                    r"[\(\)]", "", out_unit
                )  # Remove brackets but keep the content inside
                out_unit = out_unit.replace(
                    "/", "_"
                )  # Replace slashes with underscores
                out_unit = out_unit.lower()  # Convert to lowercase

            data_list.append(
                {
                    "varname": match.group("name"),
                    "Locator Type": match.group("locator_type") or None,
                    "locname": locator_name,
                    # "Branch": match.group("branch") or None,
                    # "Pipe": match.group("pipe") or None,
                    "pipename": match.group("pipe_name") or None,
                    # "Pipe Nr": match.group("pipe_nr") or None,
                    # "Pipe Number": match.group("pipe_number") or None,
                    "out_unit": out_unit,
                    "Description": match.group("description") or None,
                }
            )

        df = pd.DataFrame(data_list)
        # df = df_catalog.drop(columns=["Locator Type"])
        # df = df.drop(columns=["Pipe Nr"])
        # df = df.drop(columns=["Pipe"])
        # df = df.drop(columns=["Branch"])
        return df

    def search_catalog(self, var_name=None, loc_name=None, pipe_name=None):
        """
        Searches for variables containing a keyword in their names within a DataFrame.
        
        Arguments:
            - Var_name (str): The variable name
            - Loc_name (str): The location of the variable you want to search for
            - Pipe_name (str): the pipe name of the variable name specified located at the location name provided.
        
        Returns:
            - Pandas.DataFrame: a dataframe containing catalog information of the variables specified.

        """
        cat = self.catalog
        result_df = search(cat, var_name, loc_name, pipe_name)
        if result_df.empty:
            raise ValueError(f"We don't have {var_name} in our catalog.")
        return result_df

    def extract_trend(self, input_matrix: pd.DataFrame):
        """
        Search for variables in the DataFrame based on variable names, branches, and pipe names, and display their information.
        
        Arguments:
            - input_matrix (csv): Path to the CSV file containing variable names, branch names, and pipe names.
        
        Returns:
            - pandas.DataFrame: A DataFrame containing information for all specified variables.
        """

        self.time, self.trends, self.time_unit = self._extract_time_series_data()
        self.trends.reset_index(drop=True, inplace=True)
        df = pd.concat([self.catalog, self.trends], axis=1)
        result_dfs = []
        data = input_matrix
        for index, row in data.iterrows():
            var_name = row["varname"]
            if type(var_name) != str:
                raise ValueError(
                    "No variable name specified in row {}".format(index + 1)
                )
            loc_name = row["locname"]
            pipe_name = row["pipename"]
            row_number = row["row_number"]
            out_unit = row["out_unit"]
            time_unit = row["time_unit"]

            search_args = [var_name, loc_name, pipe_name]
            args = [v for v in search_args if type(v) == str]
            if args:
                result_df = search(df, *args)
            if result_df.empty:
                raise ValueError(
                    f"No data found for variable '{var_name}' at branch '{loc_name}'"
                )
            else:
                for _, row in result_df.iterrows():
                    unit = row["out_unit"].replace("(", "").replace(")", "").lower()
                    if pd.isna(out_unit):
                        unit = unit.replace("/", "_")
                        out_unit = unit
                    var = row["varname"]
                    unit_class = self.unitsdb["OLGA_vars"].get(var)

                    if unit_class == None:
                        for k, v in self.unitsdb["OLGA_startswith"].items():
                            if str(var).startswith(k):
                                unit_class = v
                    variable_outputs = row.filter(like="variable_output").dropna()
                    if row["Locator Type"] == "GLOBAL":
                        heading = f"{row['varname']}_{unit}"
                    else:
                        heading = f"{row['varname']}_{unit}_{row['locname']}"
                    self.time = list(dict.fromkeys(self.time))
                    if pd.notna(time_unit):
                        if self.time_unit in unit_map:
                            self.time_unit = unit_map[self.time_unit]
                            value_tagged = getattr(UnitConversion, "Time")(
                                self.time, self.time_unit
                            )
                            values = value_tagged.convert(to_unit=time_unit)
                            if time_unit in [
                                "hour",
                                "minute",
                                "second",
                                "min",
                                "s",
                                "h",
                            ]:
                                # self.time = [math.floor(value) for value in values]
                                self.time = [round(value, 2) for value in values]
                            else:
                                self.time = values
                        data = {
                            f"Time_({str(time_unit).lower()})": self.time,
                            heading: variable_outputs,
                        }
                    else:
                        time_unit == "nan"
                        time_unit = self.time_unit
                        data = {
                            f"Time_({str(self.time_unit).lower()})": self.time,
                            heading: variable_outputs,
                        }
                    trend_df = pd.DataFrame(data)
                    trend_df.set_index(f"Time_({str(time_unit).lower()})", inplace=True)
                    converted_vals = []
                    for _, row in trend_df.iterrows():
                        value = row[heading]
                        value_tagged = getattr(UnitConversion, unit_class)(value, unit)
                        conv_val = value_tagged.convert(to_unit=out_unit)
                        converted_vals.append(round(conv_val, 3))
                    trend_df.drop(columns=trend_df.columns, inplace=True)
                    trend_df[str(heading).replace(str(unit), str(out_unit))] = (
                        converted_vals
                    )
                    if pd.notna(row_number):
                        row_number = int(row_number)
                        if row_number > 0:
                            result_dfs.append(trend_df.head(row_number))
                        else:
                            result_dfs.append(trend_df.tail(-row_number))
                    else:
                        result_dfs.append(trend_df)

        if result_dfs:
            return pd.concat(result_dfs, axis=1)
        else:
            print("No data found.")
            return None

    def calc_average(
        self,
        input_matrix: pd.DataFrame,
        start_index=None,
        end_index=None,
        n_rows=None,
        n_timeunits=None,
    ):
        """
        Calculate the average of values in the DataFrame between the specified start and end indices.
        
        Arguments:
            - input_matrix (csv): The csv file containing variable names, branch names, and pipe names. This is a required argument
            - start_index (int, optional): The starting index from which the average will be calculated. Defaults to None.
            - end_index (int, optional): The ending index up to which the average will be calculated. Defaults to None.
            - n_rows (int, optional): Number of rows to consider for calculating the average. If provided, start_index and end_index are ignored. Defaults to None.
        
        Returns:
            - pandas.DataFrame: DataFrame containing the calculated averages.
        """

        data_df = self.extract_trend(input_matrix)

        if start_index is not None and end_index is not None:
            # Calculate the average between start_index and end_index
            if (
                start_index >= 0
                and end_index >= 0
                and start_index <= end_index < len(data_df)
            ):
                sliced_df = data_df.iloc[start_index - 1 : end_index]
                average = sliced_df.mean()
                return average
            else:
                raise ValueError("One of the indices is out of range")

        elif n_rows is not None:
            # Calculate the average of number of rows specified
            length = len(data_df)

            if n_rows >= 0:
                if n_rows >= length:
                    raise ValueError("Index out of range.")
                sliced_df = data_df.iloc[:n_rows]
                average = sliced_df.mean()
            else:
                n_rows = abs(n_rows)
                if n_rows > length:
                    raise ValueError("Number of rows exceeds DataFrame length.")
                last_n_values = data_df.iloc[-n_rows:]
                average = last_n_values.mean()
            return average

        elif n_timeunits is not None:

            def hours_to_seconds(hours):
                return hours * 3600

            def hours_to_minutes(hours):
                return hours * 60

            def hours_to_hours(hours):
                return hours

            def hours_to_days(hours):
                return hours / 24

            def hours_to_weeks(hours):
                return hours / 168

            def hours_to_months(hours):
                return hours / 730

            def hours_to_years(hours):
                return hours / 8760

            if n_timeunits >= 0:
                if "second" in data_df.index.name or "s" in data_df.index.name:
                    sliced_df = data_df.loc[
                        data_df.index <= hours_to_seconds(n_timeunits)
                    ]
                elif "minute" in data_df.index.name or "min" in data_df.index.name:
                    sliced_df = data_df.loc[
                        data_df.index <= hours_to_minutes(n_timeunits)
                    ]
                elif "hour" in data_df.index.name or "h" in data_df.index.name:
                    sliced_df = data_df.loc[
                        data_df.index <= hours_to_hours(n_timeunits)
                    ]
                elif "day" in data_df.index.name or "d" in data_df.index.name:
                    sliced_df = data_df.loc[data_df.index <= hours_to_days(n_timeunits)]
                elif "week" in data_df.index.name or "w" in data_df.index.name:
                    sliced_df = data_df.loc[
                        data_df.index <= hours_to_weeks(n_timeunits)
                    ]
                elif "month" in data_df.index.name or "m" in data_df.index.name:
                    sliced_df = data_df.loc[
                        data_df.index <= hours_to_months(n_timeunits)
                    ]
                elif "year" in data_df.index.name or "y" in data_df.index.name:
                    sliced_df = data_df.loc[
                        data_df.index <= hours_to_years(n_timeunits)
                    ]
                average = sliced_df.mean()

            elif n_timeunits < 0:
                total_hours = data_df.index[-1]
                end_hour = total_hours
                if "second" in data_df.index.name or "s" in data_df.index.name:
                    start_hour = total_hours + hours_to_seconds(n_timeunits)
                elif "minute" in data_df.index.name or "min" in data_df.index.name:
                    start_hour = total_hours + hours_to_minutes(n_timeunits)
                elif "hour" in data_df.index.name or "h" in data_df.index.name:
                    start_hour = total_hours + hours_to_hours(n_timeunits)
                elif "day" in data_df.index.name or "d" in data_df.index.name:
                    start_hour = total_hours + hours_to_days(n_timeunits)
                elif "week" in data_df.index.name or "w" in data_df.index.name:
                    start_hour = total_hours + hours_to_weeks(n_timeunits)
                elif "month" in data_df.index.name or "m" in data_df.index.name:
                    start_hour = total_hours + hours_to_months(n_timeunits)
                elif "year" in data_df.index.name or "y" in data_df.index.name:
                    start_hour = total_hours + hours_to_years(n_timeunits)

                sliced_df = data_df.loc[
                    (data_df.index >= start_hour) & (data_df.index <= end_hour)
                ]
                average = sliced_df.mean()

            else:
                raise ValueError("Invalid value for n_timeunits.")
            return average
        else:
            raise ValueError(
                "Invalid input provided. Please specify either start_index and end_index or n_rows along with input_matrix."
            )

    def _extract_time_series_data(self):
        """
        Process time series data from the tpl file.
        
        Arguments:
            - None
        
        Returns:
            - pandas.Series: Time information.
            - pandas.DataFrame: DataFrame containing time series variable outputs.
        """

        i = 0
        data = []
        time_series_header = None
        with open(self.filepath, "r") as file:
            for line in file:
                i += 1
                if "TIME SERIES" in line.strip():
                    match = re.search(r"TIME\sSERIES\s+'\s?\((\w)\)\s*'", line)
                    time_unit = match.group(1)
                    time_series_header = line.strip()
                elif time_series_header:
                    data.append(line.split())

        if data:
            time_unit = time_unit
            df_outputs = pd.DataFrame(data)
            df_outputs = df_outputs.T
            df_outputs = df_outputs.apply(pd.to_numeric)
            time = df_outputs.iloc[0]
            df_outputs = df_outputs.iloc[1:]
            df_outputs.columns = ["variable_output"] * len(df_outputs.columns)
            return time, df_outputs, time_unit
        else:
            return pd.Series(), pd.DataFrame()


class tplBatchParser:
    """
    Class to handle batches of tpl files
    
    Functions:
        1. extract_trends: to extract trends from a list of tpl files
        2. calc_averages: to compute averages of trends extracted from a list of tpl files
    """
    def __init__(self, list_of_files):
        """
        Initializes an instance of the class by loading file data and unit definitions.
        
        Arguments:
            - filepaths (list): a list of the path to the ppl files to be loaded.
        Returns:
            - None
        """

        self.list_of_files = list_of_files
        self.files = [tplParser(file) for file in list_of_files]

    def extract_trends(self, input_matrix: pd.DataFrame):
        """
        Function to extract trends from a batch of tpl files
        
        Arguments:
            - Input_matrix (csv): The matrix containing variable names, branch names, and pipe names. This is a required argument
        
        Returns
            - pandas.DataFrame: DataFrame containing extracted trends from tpl files
        
        """

        # create suffixes for distinguishing between columns
        suffixes = ["_" + str(os.path.basename(file)) for file in self.list_of_files]
        list_of_dfs = [file.extract_trend(input_matrix) for file in self.files]
        # attaching suffixes to dataframes
        for i in range(len(list_of_dfs)):
            list_of_dfs[i] = list_of_dfs[i].add_suffix(suffixes[i])
        # merging final dataframe
        final_df = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True, how="outer"),
            list_of_dfs,
        )
        return final_df

    def calc_averages(
        self,
        input_matrix: pd.DataFrame,
        start_index=None,
        end_index=None,
        n_rows=None,
        n_timeunits=None,
    ):
        """
        Calculate the average of values in the DataFrame up to the specified index or of the last n values.

        Arguments:
            - input_matrix (csv): The csv file containing variable names, branch names, and pipe names. This is a required argument
            - start_index (int, optional): The starting index from which the average will be calculated. Defaults to None.
            - end_index (int, optional): The ending index up to which the average will be calculated. Defaults to None.
            - n_rows (int, optional): Number of rows to consider for calculating the average. If provided, start_index and end_index are ignored. Defaults to None.

        Returns:
            - pandas.DataFrame: DataFrame containing the calculated averages.
        """
        suffixes = ["_" + str(os.path.basename(file)) for file in self.list_of_files]
        list_of_dfs = [
            file.calc_average(input_matrix, start_index, end_index, n_rows, n_timeunits)
            .to_frame()
            .T
            for file in self.files
        ]
        for i in range(len(list_of_dfs)):
            list_of_dfs[i] = list_of_dfs[i].add_suffix(suffixes[i])

        final_df = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True, how="outer"),
            list_of_dfs,
        )
        final_df = final_df.T
        final_df.columns = ["Average"]
        return final_df


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="parse arguments for obtaining variables from catalog"
    )

    parser.add_argument("-f", "--filepath", type=str, help="enter the path to tpl file")
    parser.add_argument(
        "-v",
        "--varname",
        type=str,
        help="enter the name of the tpl variable",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-l",
        "--loc",
        type=str,
        help="enter the name of the tpl branch",
        required=False,
        default=None,  # to ensure all variable time series output is pulled if no branch is provided
    )
    parser.add_argument(
        "-p",
        "--pipe",
        type=str,
        help="enter the name of the tpl pipe",
        required=False,
        default=None,  # to ensure all variable time series output is pulled if no pipe is provided
    )

    parser.add_argument(
        "-is",
        "--start_index",
        type=int,
        help="enter the start index to calculate average for trend",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-ie",
        "--end_index",
        type=int,
        help="enter the last index to calculate average for trend",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-n",
        "--n_rows",
        type=int,
        help="enter the number of rows to calculate average for trend",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-nt",
        "--n_timeunits",
        type=int,
        help="enter the number of hours to calculate average for trend",
        required=False,
        default=None,
    )

    parser.add_argument(
        "-c",
        "--csv_file",
        type=str,
        help="Path to the CSV file containing variable names, branch names, and pipe names",
    )

    args = parser.parse_args()
    args.filepath = args.filepath.replace("\\", "/")
    tplparser = tplParser(args.filepath)
    input_matrix = pd.read_csv(args.csv_file)
    data = tplparser.extract_trend(input_matrix)
    # data1 = tplparser.calc_average(
    #     input_matrix=input_matrix, n_timeunits=args.n_timeunits
    # )
    # data2 = tplparser.search_catalog(var_name=args.varname, pipe_name=args.pipe)
    print(data)



"""
PPL parser functionality
"""

import warnings
import cProfile
import pstats
import time as t
# temporary to remove pandas 3.0 deprecationwarning about pyarrow from printing
warnings.filterwarnings("ignore", category=DeprecationWarning)

import pandas as pd
import numpy as np
import argparse
import re
from functools import reduce
from pathlib import Path
import os
import sys
import math
import yaml
from io import StringIO


sys.path.insert(0, "src")
from pypontem.utils.unit_conversion import UnitConversion, unit_map

current_dir = os.path.dirname(os.path.abspath(__file__))
YAML_PATH = os.path.join(current_dir, "..", "utils", "units.yaml")

# YAML_PATH = "src/pypontem/utils/units.yaml"


def read_file(file_path):
    if os.path.isfile(file_path):
        with open(file_path, "r") as file:
            content = file.read()
        return content
    else:
        raise FileNotFoundError(
            f"The file specified '{file_path}' does not exist. Please verify the location of the file."
        )


def open_yaml(yaml_path):
    with open(yaml_path, "r") as file:
        unitsdb = yaml.safe_load(file)
    return unitsdb


pd.set_option("display.max_rows", None)
pd.set_option("display.max_columns", None)
pd.set_option("display.max_colwidth", None)


def search(df, var_name=None, loc_name=None, pipe_name=None):
    r"""
    Searches for variables containing a keyword in their names within a DataFrame.
        Args:
            df (pandas.DataFrame): The DataFrame to search within.
            keyword (str): The keyword to search for.

        Returns:
            None
    """
    filter_conditions = []
    if var_name:
        var_name = var_name.upper()
        filter_conditions.append(df["Name"] == var_name)
    if loc_name:
        loc_name = loc_name.upper()
        filter_conditions.append(df["Locator Name"] == loc_name)
    if pipe_name:
        pipe_name = pipe_name.upper()
        filter_conditions.append(
            df["Pipe Name"].str.contains(pipe_name, na=False, case=False)
        )
    if len(filter_conditions) > 1:
        result_df = df[reduce(lambda x, y: x & y, filter_conditions)]
    else:
        result_df = df[filter_conditions[0]]
    return result_df


class pplParser:
    """
    Class which holds the ppl parser functions
    
    Functions:
        1. metadata: to extract metadata from a ppl file
        2.branch_names: to extract branch names from a ppl file
        3.branch_profiles: to extract branch profile information from a ppl file
        4.parse_number_of_variables: to extract the number of variables present in a ppl file
        5. extract_catalog: to extract 
        6. search_catalog: to extract the information from the catalog of the variable specified
        7. extract_trends: to extract trends in a ppl file
        8. extract_trend_join_nodes: to join nodes of branches extracted in a ppl file	
    """

    def __init__(self, filepath):
        """
        Initializes an instance of the class by loading file data and unit definitions.
        
        Arguments:
            - filepath (str): The path to the ppl file to be loaded.
        """


        # read file from path
        self.filepath = Path(filepath)
        # save content of file in memory so we don't open it repeatedly - NOTE: may need to be changed later
        self.content = read_file(self.filepath)
        # open unit yaml file
        self.unitsdb = open_yaml(YAML_PATH)

    @property
    def metadata(self):
        """
        Extracts metadata from the content stored in the object and returns it as a pandas DataFrame.
        
        Arguments:
            - file_path (str): the path to the ppl file
        
        Returns:
            - pandas.DataFrame: A DataFrame containing the extracted metadata with keys as column headers
        """
        return self._extract_metadata()

    def _extract_metadata(self):
        metadata = {}  # Dictionary to store metadata

        try:
            branch_index = self.content.index("BRANCH\n")
        except ValueError:
            branch_index = len(self.content)

        # Extract metadata before "BRANCH"
        metadata_lines = self.content[:branch_index]

        # Parsing specific metadata and storing in dictionary
        for i, line in enumerate(metadata_lines.splitlines()):
            line = line.strip()
            olga_match = re.search(r"OLGA\s+(\S+)", line)
            if olga_match:
                metadata["version"] = line.strip("'")
            elif "INPUT FILE" in line:
                metadata["Input file"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "PVT FILE" in line:
                metadata["pvt"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "DATE" in line:
                metadata["time"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "PROJECT" in line:
                metadata["project"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "TITLE" in line:
                metadata["title"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "AUTHOR" in line:
                metadata["author"] = [
                    metadata_lines.splitlines()[i + 1].strip().strip("'")
                ]
            elif "GEOMETRY" in line:
                match = re.search(r"\((.*?)\)", line)
                if match:
                    metadata["geometry"] = [match.group(1)]
        return pd.DataFrame(metadata)
    
    @property
    def branch_names(self):
        """
        Extracts and prints the names of all branches in the given ppl file.
        
        Arguments:
            - None.
        
        Returns:
            - branch_names (list): A list of extracted branch names as strings.
        """

        return self._extract_branch_names()

    def _extract_branch_names(self):
        df_branch_names = []
        pattern = "(BRANCH|ANNULUS)\n(.+)"
        branch_names = re.findall(pattern, self.content)
        for branch in branch_names:
            df_branch_names.append(
                branch[1]
            )  # Accessing the second element (branch name) of each tuple
        return df_branch_names

    @property
    def branch_profiles(self):
        """
        Extracts and displays elevation for a specific branch or all branches in the file.
        
        Arguments:
            - target_branch (str, optional): The specific branch to display data for. Defaults to None.
        
        Returns:
            - pandas.Dataframe: a dataframe containing the profiles of the specified branch.
        """

        return self._extract_branch_profiles(target_branch=None)

    def _extract_branch_profiles(self, target_branch=None):
        def create_branch_table(branch_data):
            # Ensure lengths and elevations have the same length
            min_length = min(
                len(branch_data["Lengths"]), len(branch_data["Elevations"])
            )
            df = pd.DataFrame(
                {
                    "Lengths_("
                    + str(self.metadata["geometry"].values[0]).lower()
                    + ")": branch_data["Lengths"][:min_length],
                    "Elevations_("
                    + str(self.metadata["geometry"].values[0]).lower()
                    + ")": branch_data["Elevations"][:min_length],
                    "Pipe Number":num_pipes,
                }
            )
            return df

        # Define a pattern to match each 'BRANCH' section
        branch_pattern = re.compile(
            r"(BRANCH|ANNULUS)\n'(.*?)'\n(\d+)\n([\s\S]*?)(?=(BRANCH|ANNULUS)|\nCATALOG)",
            re.DOTALL,
        )

        # Find all matches
        matches = branch_pattern.findall(self.content)
        found = False  # Flag to check if the target branch is found

        branches = {}

        for match in matches:
            branch_data = {}
            branch_name = match[1]
            num_pipes = int(match[2])
            data_section = match[3].split()
            if data_section:
                midpoint = len(data_section) // 2
                lengths = [float(data) for data in data_section[:midpoint]]
                elevations = [float(data) for data in data_section[midpoint:]]
                branch_data = {
                    "Branch": branch_name,
                    "Lengths": lengths,
                    "Elevations": elevations,
                }

                # Display branch data only if it matches the target branch
                if not target_branch or branch_name == target_branch:
                    found = True  # Set found to True if the branch is found
                    profile = create_branch_table(branch_data)
                    branches[branch_name] = profile

        if (
            not found and target_branch
        ):  # Display message if branch not found and target_branch is provided
            raise ValueError(
                f"The branch '{target_branch}' does not exist in the file."
            )
        else:
            return branches
        
    @property
    def n_vars(self):
        """
        Parses the number of variables from a given file path.
        
        Arguments:
            - None
        
        Returns:
            - n_vars (int): The number of variables if found, else None.
        """

        return self._parse_number_of_variables()

    def _parse_number_of_variables(self):
        # regex pattern which finds the next line after CATALOG
        pattern = re.search(r"CATALOG[\r \n]+([^\r\n]+)", self.content)
        return int(pattern[1])

    @property
    def catalog(self):
        """
        Extract variable information from the provided content using a regular expression pattern.
        
        Arguments:
            - None.
        
        Returns:
            - pandas.DataFrame: A DataFrame containing all the catalog information available in the ppl file.

        """
        return self._extract_catalog()

    def _extract_catalog(self):
        pattern = re.compile(
            r"""
            (?P<name>\S+)                      # variable name
            \s+'(?P<locator_type>[^']*)'       # locator type
            (\s+'(?P<locator_name>[^']*)')?    # locator name (optional)
            (\s+'(?P<branch>[^']*)')?          # branch (optional)
            (\s+'(?P<pipe>[^']*)')?            # pipe (optional)
            (\s+'(?P<pipe_name>[^']*)')?       # pipe name (optional)
            (\s+'(?P<pipe_nr>[^']*)')?         # pipe nr (optional)
            (\s+'(?P<pipe_number>[^']*)')?     # pipe number (optional)
            \s+'(?P<unit>[^']*)'               # unit
            \s+'(?P<description>[^']*)'        # description
            """,
            re.IGNORECASE | re.VERBOSE,
        )
        
        
        lines = self.content.splitlines()
        catalog_index = next(
            (
                i + 1
                for i, line in enumerate(lines)
                if line.strip() == "CATALOG"
                and i + 1 < len(lines)
                and lines[i + 1].strip().isdigit()
            ),
            None,
        )
        # Extract variable information from the lines after the "CATALOG" line
        matches = pattern.finditer("\n".join(lines[catalog_index:]))
        data_list = []
        # Iterate over matches and extract data
        for match in matches:
            locator_name = match.group("locator_name") or None
            if locator_name == "BRANCH:":
                locator_name = match.group("branch") or None
            data_list.append(
                {
                    "Name": match.group("name"),
                    "Locator Type": match.group("locator_type") or None,
                    "Locator Name": locator_name,
                    # "Branch": match.group("branch") or None,
                    #"Pipe": match.group("pipe") or None,
                    "Pipe Name": match.group("pipe_name") or None,
                    #"Pipe Nr": match.group("pipe_nr") or None,
                    "Pipe Number": match.group("pipe_number") or None,
                    "Unit": match.group("unit") or None,
                    "Description": match.group("description") or None,
                }
            )

        df = pd.DataFrame(data_list)
        # df = df.drop(columns=["Pipe Nr"])
        # df = df.drop(columns=["Pipe"])
        # df = df.drop(columns=["Branch"])
        return df
    
    def search_catalog(self, var_name=None, loc_name=None, pipe_name=None):
        """
        Searches for variables containing a keyword in their names within a DataFrame.
        
        Arguments:
	        - Var_name (str): The variable name
	        - Loc_name (str): The location of the variable you want to search for
	        - Pipe_name (str): the pipe name of the variable name specified located at the location
        
        Returns:
            - pandas.DataFrame: a dataframe containing catalog information of the variable specified at the location and pipe name provided.

        """
        cat = self.catalog
        result_df = search(cat, var_name, loc_name, pipe_name)
        if result_df.empty:
            raise ValueError(f"We don't have {var_name} in our catalog.")
        return result_df
    
    def _extract_time_series_data(self):
        """
        Process time series data from the ppl file.
        
        Arguments:
            - None
        
        Returns:
            - pandas.DataFrame: DataFrame containing time series variable outputs.
        """

        # Define regex pattern to extract the number of rows from CATALOG section
        data = self.content
        catalog_pattern = r'CATALOG\s+(\d+)'
        catalog_match = re.search(catalog_pattern, data)

        if catalog_match:
            number_of_rows = int(catalog_match.group(1))
        else:
            raise ValueError("Number of rows under CATALOG not found")

        # Define regex pattern to extract the TIME SERIES section
        time_series_pattern = r'TIME SERIES\s+\'\s*\(S\)\s*\'([\s\S]+?)(?=\n\n|\Z)'
        time_series_match = re.search(time_series_pattern, data)
        if time_series_match:
            time_series_section = time_series_match.group(0)
            bracket_match = re.search(r"\((.*?)\)", time_series_section)
            time_unit = bracket_match.group(1) if bracket_match else None
            time_series_data = time_series_match.group(1).strip()
        else:
            raise ValueError("TIME SERIES section not found")
        content = self._extract_catalog()
        variable_Names = content["Name"].reset_index(drop=True)
        # variable_Names.reset_index(drop=True)
        branch_names = content["Locator Name"].reset_index(drop=True)
        # branch_names.reset_index(drop=True)
        units = content["Unit"].reset_index(drop=True)
        # units.reset_index(drop=True)
        # Process the extracted time series data
        lines = time_series_data.split('\n')
        time_dict = {}

        for line in lines:
            values = line.split()
            if len(values) == 1:
                current_time = values[0]
                time_dict[current_time] = []
            else:
                time_dict[current_time].append(values[0:])

        time_out = []
        values_out = []
        for time, values in time_dict.items():
            time_out.append(time)
            values_out.append(values)
        
        columns = []
        data = []

        for idx, time in enumerate(time_out):
            for var_name, branch_name,unit in zip(variable_Names, branch_names,units):
                columns.append(f"{branch_name}_{var_name}_time_in_{time_unit}:_{time}_{unit}")
            for row in values_out[idx]:
                data.append(row)

         # print(data)
        # Checking that the number of columns matches the number of data points
        expected_columns = len(variable_Names) * len(time_out)
        if len(columns) != expected_columns:
            raise ValueError(f"Expected {expected_columns} columns, but got {len(columns)}")
        # Converting the  data to the correct format
        
        flattened_data = []
        for values in values_out:
            for value_set in values:
                flattened_data.append(value_set)
        if len(flattened_data) % len(columns) != 0:
            raise ValueError(f"Data length {len(flattened_data)} is not a multiple of the number of columns {len(columns)}")
         
        df = pd.DataFrame([flattened_data[i:i + len(columns)] for i in range(0, len(flattened_data), len(columns))], columns=columns)
        df1= df.values
        flattened_data = []
        max_length = max(len(sublist) for sublist in df1)
        max_length = max(len(sublist) for sublist in data)

        flattened_data_np = np.full((len(data), max_length), None, dtype=object)
        for i, sublist in enumerate(data):
            flattened_data_np[i, :len(sublist)] = sublist
        transposed_data = flattened_data_np.T

        df_final = pd.DataFrame(transposed_data, columns=columns)
        return df_final

    def extract_trend(
            self, 
            input_matrix: pd.DataFrame 
            ):
        """
        Extracts and processes trend data from an input matrix, performing unit conversions and time filtering.
        
        Arguments:
            - input_matrix (csv): A csv file containing input data, including variable names, branches, units, and time specifications.
        
        Returns:
            - pandas.DataFrame: A combined DataFrame containing the processed trend data with converted units and filtered time ranges.
        """

        # calling the output of catalog, profiles, metadata and time series methods
        catalog = self._extract_catalog()
        profiles = self.branch_profiles
        metadata = self.metadata
        profile_unit = metadata['geometry']
        profile_unit_cleaned = profile_unit.str.lower()
        time_series = self._extract_time_series_data()
        df_catalog = catalog.filter(items=['Name','Locator Type','Locator Name','Unit'])
        df = pd.concat(profiles.values(), keys=profiles.keys()).reset_index(level=0).rename(columns={'level_0': 'Category'})
        df.drop(columns=['Elevations_(m)'], inplace=True)
        data_file = input_matrix
        data_file = data_file.dropna(how="all")
        variable_names = data_file['varname'].to_list()
        trend_df =[]
        for index, row in data_file.iterrows():
            var_name = row["varname"]
            if type(var_name) != str:
                raise ValueError(
                    "No variable name specified in row {}".format(index + 1)
                )
            branch_name = row["branchname"]
            out_unit = row["out_unit"]
            out_unit_profile = row["out_unit_profile"]
            out_time_unit = row["time_unit"]
            start_time = row["start_time"]
            end_time = row["end_time"]
            data = time_series.filter(regex=f'^{branch_name}_{var_name}_')
            match = df_catalog[(df_catalog['Name'] == var_name) & (df_catalog['Locator Name'] == branch_name)]
            match.reset_index(drop=True, inplace=True)
            unit = match['Unit']
            unit_cleaned = unit.str.extract(r'\((.*?)\)', expand=False).str.lower()
            if not match.empty:
                location =  match['Locator Type'].values[0]
            else:
                location = None
            columns = []
            data1 = []

            if location is not None:
                pipe_data = df[df['Category'] == branch_name]
                columns.append(branch_name)
                if location == "BOUNDARY:":
                    data1.extend(pipe_data['Lengths_(m)'].tolist())
                    result = pd.DataFrame(data1, columns=['Profiles'])
                else:
                    data1.extend(pipe_data['Lengths_(m)'].tolist())
                    result_list = [(data1[i] + data1[i + 1]) / 2 for i in range(len(data1) - 1)]
                    result = pd.DataFrame(result_list, columns=['Profiles'])   
            else:
                raise ValueError (f"Branch(es) given in the input matrix are not available in the catalog, double check your Branch inputs")
                
            
        # Converting Profiles values using unit conversion class       
            con_vals = []
            for i in result['Profiles']:
                unit = profile_unit_cleaned
                unit = unit.to_string(index=False)
                if pd.isna(out_unit_profile):
                    out_unit_profile = unit

                value = i
                value = float(value)
            
                value_tagged = getattr(UnitConversion, "Length")(value, unit)
                conv_val = value_tagged.convert(to_unit=out_unit_profile)
                con_vals.append(round(conv_val, 3))
            col_name = f"Profiles_{branch_name}_{var_name}_{out_unit_profile}"
            result1 = pd.DataFrame(con_vals, columns=[col_name])
            # End of converting profiles values
            # Starting to prep and convert trend and time values
            final_df = data.dropna()
            unit = unit_cleaned
            unit = unit.to_string(index=False)
            var = var_name
            unit_class = self.unitsdb["OLGA_vars"].get(var)
            converted_vals = []
            cols = final_df.columns
            #Converting time values from the trend column names
            time_vals = []
            updated_column_vals =[]
            for column in cols:
                time_values = re.findall(r"[-+]?\d*\.\d+e[-+]?\d+|[-+]?\d+\.\d+", column)
                for val in time_values:
                    values = float(val)
                    time_unit_match = re.search(r"in_(\w+):", column)
                    time_unit_init = time_unit_match.group(1)
                    if time_unit_init in unit_map:
                        time_unit = unit_map[time_unit_init]

                    if pd.isna(out_time_unit):
                        out_time_unit = time_unit

                    value_tagged = getattr(UnitConversion, "Time")(values, time_unit)
                    conv_val = value_tagged.convert(to_unit=out_time_unit)
                    time_vals.append(round(conv_val, 3))
                    for time_val in time_vals:
                        updated_text = column.replace(val, f"{time_val}")
                        # updated_text1 = updated_text.replace(time_unit_init, out_time_unit)
                        updated_text1 = re.sub(r'(_in_)' + time_unit_init + r'(:)', r'\1' + out_time_unit + r'\2', updated_text)

                    updated_column_vals.append(updated_text1)
            # Update the column names with the converted time values
            final_df.columns = updated_column_vals
            # Filter based on time range
            # Extract numeric time values from column names
            time_numeric = final_df.columns.str.extract(r'([-+]?\d*\.\d+|\d+)', expand=False).astype(float)

            # Convert to a Series to use .between()
            time_values_series = pd.Series(time_numeric.values.flatten(), index=final_df.columns)

            # Checking if start_time and end_time are specified
            if pd.notna(start_time) and pd.notna(end_time):
                if start_time > end_time:
                    raise ValueError(f"The start time cannot be greater than the end time. Please ensure the time range is valid and try again")
                else:
                    valid_time_mask = time_values_series.between(start_time, end_time)
                    filtered_columns = final_df.columns[valid_time_mask]
            elif pd.notna(end_time):
                raise ValueError(f"The start time is not specified. Please provide a valid start time to proceed")
            elif pd.notna(start_time):
                raise ValueError(f"The end time is not specified. Please provide a valid end time to proceed")
            else:
                filtered_columns = final_df.columns

            filtered_df = final_df[filtered_columns]
            column_names = filtered_df.columns
            # Convert values in the filtered DataFrame
            for _, row in filtered_df.iterrows():
                for column_name in filtered_df.columns:
                    value = float(row[column_name])
                    if pd.isna(out_unit):
                        out_unit = unit

                    value_tagged = getattr(UnitConversion, unit_class)(value, unit)
                    conv_val = value_tagged.convert(to_unit=out_unit)
                    converted_vals.append(round(conv_val, 3))
            reshaped_values = [converted_vals[i:i + len(column_names)] for i in range(0, len(converted_vals), len(column_names))]
            # Create a DataFrame from the reshaped list
            structured_df = pd.DataFrame(reshaped_values, columns=column_names)
            # Define a function to update the part within brackets
            def update_brackets(column_name, new_value):
                # Use regular expression to replace the part within brackets
                updated_name = re.sub(r'\(.*\)', f'({new_value})', column_name)
                return updated_name
            # Apply the function to each column name
            new_column_names = [update_brackets(cola, out_unit) for cola in structured_df.columns]
            # Assign the new column names to the DataFrame
            structured_df.columns = new_column_names
            trends = pd.concat([result1, structured_df], axis=1)
            trends.set_index(col_name)
            trend_df.append(trends)

        combined_df = pd.concat(trend_df, axis=1, keys=variable_names, ignore_index=False)
        # Sort each column to push NaNs to the bottom
        for col in combined_df.columns:
            combined_df[col] = combined_df[col].dropna().tolist() + [np.nan] * combined_df[col].isna().sum()
        # combined_df.to_csv('combined_df1.csv', index=True)
        return combined_df
        

    def extract_trends_join_nodes(
            self,
            input_matrix: pd.DataFrame,
            branch_matrix: pd.DataFrame,
    ):
        """ 
        Extracts and processes trend data for branches, combining boundary and section data. 
        
        Arguments:
            - input_matrix (csv): A csv file containing trend data for various branches with columns ['branchname', 'varname', 'out_unit', 'out_unit_profile', 'time_unit']. 
            - branch_matrix (csv): A csv file containing information about branch connections with columns ['branch_in', 'branch_out']. 
        
        Returns: 
            - pandas.DataFrame: A combined DataFrame containing processed trend data for the specified branches, including boundary and section data, with consistent units and profiles. 
        """

        data_df = self.extract_trend(input_matrix) 
        catalog = self._extract_catalog()
        df_catalog = catalog.filter(items=['Name','Locator Type','Locator Name'])
        branch_profiles = self.branch_profiles
        result = {}
        for section, df in branch_profiles.items():
            result[section] = df['Pipe Number'].unique()

        # Convert the result to a DataFrame for better visualization
        result_df = pd.DataFrame.from_dict(result)
        # Select one level of the MultiIndex (e.g., level 1)
        data_df.columns = data_df.columns.get_level_values(1)

        branch_matrix['branch_in'] = branch_matrix['branch_in'].str.strip().str.upper()
        branch_matrix['branch_out'] = branch_matrix['branch_out'].str.strip().str.upper()
        input_branches = input_matrix['branchname'].str.strip().str.upper()
        catalog_branches = set(df_catalog['Locator Name'].str.strip().str.upper())

        # Check for branches in the branch_matrix that are not in the catalog or input matrix
        for branch in branch_matrix['branch_in'].unique():
            if branch not in catalog_branches:
                raise ValueError(f"Branch in '{branch}' in branch_matrix is not present in the catalog.")
            elif branch not in input_branches.unique():
                raise ValueError(f"Branch in '{branch}' in branch_matrix is not present in the input_matrix.")
        for branch in branch_matrix['branch_out'].unique():
            if branch not in catalog_branches:
                raise ValueError(f"Branch out '{branch}' in branch_matrix is not present in the catalog.")
            elif branch not in input_branches.unique():
                raise ValueError(f"Branch out '{branch}' in branch_matrix is not present in the input_matrix.")
        
        data_file = input_matrix
        data = branch_matrix
        first_branch = data["branch_in"]
        num_of_pipes = result_df[first_branch]
        #num = num_of_pipes.item()
        num = num_of_pipes.values.flatten()
        df_boundary = []
        df_section = []
        for index, row in data.iterrows():
            branch_in = row['branch_in']
            branch_out = row['branch_out'] 
            variable_names = data_file['varname'].unique()
            for v in variable_names:
                match = df_catalog[(df_catalog['Name'] == v) & (df_catalog['Locator Name'] == branch_in) | (df_catalog['Name'] == v) & (df_catalog['Locator Name'] == branch_out)]
                match.reset_index(drop=True, inplace=True)

                if match.empty:
                    raise ValueError(f"No matching catalog entry found for branch_in '{branch_in}' or branch_out '{branch_out}' with variable '{v}'.")

                location =  match['Locator Type'].values[0]
                
                if location == "BOUNDARY:":
            
                    branch_in_cols = [col for col in data_df.columns if branch_in in col and v in col]
                    branch_out_cols = [col for col in data_df.columns if branch_out in col and v in col]
                    # Separating dataframe based on branch_in and Branch_out
                    # Identifying overlapping columns 
                    overlapping_cols = list(set(branch_in_cols) & set(branch_out_cols))
                    if not overlapping_cols:
                        branch_in_cols = [col for col in data_df.columns if branch_in in col and v in col]
                        branch_out_cols = [col for col in data_df.columns if branch_out in col and v in col] 
                        # Separating dataframe based on branch_in and Branch_out
                        # print(branch_in_cols)
                        df_branch_in = data_df[branch_in_cols]
                        df_branch_out = data_df[branch_out_cols]
                    else:
                        # Removing overlaps from one list and keeping them in the other
                        # Keeping overlapping columns in branch_in only
                        branch_in_cols = list(set(branch_in_cols))  # Deduplicate within each list
                        branch_out_cols = list(set(branch_out_cols) - set(overlapping_cols))
                        # Separating dataframe based on branch_in and branch_out
                        df_branch_in = data_df[branch_in_cols + overlapping_cols]  # Include overlaps here only once
                        df_branch_out = data_df[branch_out_cols]

                    def rename_columns(col):
                        profiles_match = re.match(r'Profiles_.*_(\w+)$', col)
                        if profiles_match:
                            return f'Profiles_{v}_{profiles_match.group(1)}'
                        
                        pattern = r"in_(.*?):"
                        match = re.search(pattern, col)
                        variable = match.group(1) 
                        m = re.match(rf'.*_(time_in_{variable}:_\d+\.\d+_\(.*?\))', col)
                        if m:
                            return m.groups()[0]
                        return col  
                        
                    # Applying the renaming function to both columns of both branch_in and Branch_out
                    df_branch_in.columns = [rename_columns(col) for col in df_branch_in.columns]
                    df_branch_out.columns = [rename_columns(col) for col in df_branch_out.columns]
                    df_boundary.append(df_branch_in)
                    df_boundary.append(df_branch_out)
                    num_rows_to_process = [i+1 for i in num]
                    # print(result_df)
                    combined_boundary_df = pd.concat(df_boundary)  
                    combined_boundary_df = combined_boundary_df.dropna()
                    prof = combined_boundary_df.filter(like="Profiles").copy()

                    for i in range(0, len(num_rows_to_process), 2):
                        start_index = num_rows_to_process[i]
                        step_count = num_rows_to_process[i + 1]

                        last_value = prof.iloc[start_index - 1] 
                        end_index = min(start_index + step_count, len(prof))
                        prof.iloc[start_index:end_index] += last_value 
                        last_value_added = prof.iloc[end_index - 1]
                        if end_index < len(prof):
                            prof.iloc[end_index:] += last_value_added
                    combined_boundary_df[prof.columns] = prof

                
                else:
                    branch_in_cols = [col for col in data_df.columns if branch_in in col and v in col]
                    branch_out_cols = [col for col in data_df.columns if branch_out in col and v in col]
                    # Separating dataframe based on branch_in and Branch_out
                    # Identifying overlapping columns 
                    overlapping_cols = list(set(branch_in_cols) & set(branch_out_cols))
                    if not overlapping_cols:
                        branch_in_cols = [col for col in data_df.columns if branch_in in col and v in col]
                        branch_out_cols = [col for col in data_df.columns if branch_out in col and v in col] 
                        # Separating dataframe based on branch_in and Branch_out
                        # print(branch_in_cols)
                        df_branch_in = data_df[branch_in_cols]
                        df_branch_out = data_df[branch_out_cols]
                    else:
                        # Removing overlaps from one list and keeping them in the other
                        # Keeping overlapping columns in branch_in only
                        branch_in_cols = list(set(branch_in_cols))  # Deduplicate within each list
                        branch_out_cols = list(set(branch_out_cols) - set(overlapping_cols))
                        # Separating dataframe based on branch_in and branch_out
                        df_branch_in = data_df[branch_in_cols + overlapping_cols]  # Include overlaps here only once
                        df_branch_out = data_df[branch_out_cols]
                    # Function to automate renaming based on regex patterns
                    def rename_columns(col):
                        profiles_match = re.match(r'Profiles_.*_(\w+)$', col)
                        if profiles_match:
                            return f'Profiles_{v}_{profiles_match.group(1)}'
                        
                        pattern = r"in_(.*?):"
                        # Search for the pattern in the text
                        match = re.search(pattern, col)
                        variable = match.group(1)
                        m = re.match(rf'.*_(time_in_{variable}:_\d+\.\d+_\(.*?\))', col)
                        if m:
                            return m.groups()[0]
                        return col  
                        
                    # Applying the renaming function to both columns of both branch_in and Branch_out
                    df_branch_in.columns = [rename_columns(col) for col in df_branch_in.columns]
                    df_branch_out.columns = [rename_columns(col) for col in df_branch_out.columns]
                    df_section.append(df_branch_in)
                    df_section.append(df_branch_out)
                    
                    combined_section_df = pd.concat(df_section)   
                    combined_section_df = combined_section_df.drop_duplicates().dropna()
                    prof_section = combined_section_df.filter(like="Profiles").copy()
                    
                    for i in range(0, len(num), 2):
                        start_index = num[i]
                        step_count = num[i + 1]

                        last_value = prof_section.iloc[start_index - 1] 
                        end_index = min(start_index + step_count, len(prof_section))
                        prof_section.iloc[start_index:end_index] += last_value 
                        last_value_added = prof_section.iloc[end_index - 1]
                        if end_index < len(prof_section):
                            prof_section.iloc[end_index:] += last_value_added
                    combined_section_df[prof_section.columns] = prof_section

        if len(df_boundary) == 0:                
            combined_section_df = combined_section_df.reset_index(drop=True)
            return combined_section_df
        
        elif len(df_section) == 0:
            combined_boundary_df = combined_boundary_df.reset_index(drop=True)
            return combined_boundary_df
        
        else:
            combined_boundary_df = combined_boundary_df.reset_index(drop=True)
            combined_section_df = combined_section_df.reset_index(drop=True)
            max_len = max(len(combined_boundary_df), len(combined_section_df))

            combined_boundary_df = combined_boundary_df.reindex(range(max_len))
            combined_section_df = combined_section_df.reindex(range(max_len))

            final_df= pd.concat([combined_boundary_df,combined_section_df], axis=1, keys=variable_names, ignore_index=False)
        return final_df
            
    
        
class pplBatchParser:
    """
    Class to handle batches of ppl files
    
    Functions:
        1. extract_trends: to extract trends from a list of ppl files
        2. join_batch_nodes: to join nodes of branches extracted from a list of ppl files
    """


    def __init__(self, list_of_files):
        """
        Initializes an instance of the class by loading file data and unit definitions.
        
        Arguments:
            - filepaths (list): a list of the path to the ppl files to be loaded.
        Returns:
            - None
        """

        self.list_of_files = list_of_files
        self.files = [pplParser(file) for file in list_of_files]

    def extract_trends(self, input_matrix: pd.DataFrame):
        """
        Function to extract trends from a batch of ppl files
        
        Arguments:
            - Input_matrix (csv): A csv file containing variable names, branch names, and pipe names. This is a required argument.

        Returns
            - pandas.DataFrame: DataFrame containing extracted trends from ppl files
        """

        # create suffixes for distinguishing between columns
        suffixes = ["_" + str(os.path.basename(file)) for file in self.list_of_files]
        list_of_dfs = [file.extract_trend(input_matrix) for file in self.files]
        # attaching suffixes to dataframes
        for i in range(len(list_of_dfs)):
            list_of_dfs[i] = list_of_dfs[i].add_suffix(suffixes[i])
        # merging final dataframe
        final_df = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True, how="outer"),
            list_of_dfs,
        )
        return final_df

    def join_batch_nodes(self, input_matrix: pd.DataFrame, branch_matrix: pd.DataFrame):
        """
        Extracts and processes trend data for branches, combining boundary and section data from a list of ppl files. 
        
        Arguments:
            - input_matrix (csv): A csv file containing trend data for various branches with columns ['branchname', 'varname', 'out_unit', 'out_unit_profile', 'time_unit']. 
            - branch_matrix (csv): A csv file containing information about branch connections with columns ['branch_in', 'branch_out']. 

        Returns: 
            - pandas.DataFrame: A combined DataFrame containing processed trend data for the specified branches, including boundary and section data, with consistent units and profiles. 
        """

        # create suffixes for distinguishing between columns
        suffixes = ["_" + str(os.path.basename(file)) for file in self.list_of_files]
        list_of_dfs = [file.extract_trends_join_nodes(input_matrix, branch_matrix) for file in self.files]
        # attaching suffixes to dataframes
        for i in range(len(list_of_dfs)):
            list_of_dfs[i] = list_of_dfs[i].add_suffix(suffixes[i])
        # merging final dataframe
        final_df = reduce(
            lambda x, y: pd.merge(x, y, left_index=True, right_index=True, how="outer"),
            list_of_dfs,
        )
        return final_df
           

if __name__ == "__main__":
    with cProfile.Profile() as profile:
        parser = argparse.ArgumentParser(
            description="parse arguments for obtaining variables from catalog"
        )

        parser.add_argument("-f", "--filepath", type=str, help="enter the path to tpl file")
        #parser.add_argument("-f", "--filepath", type=str,nargs='+', help="enter the path to tpl file")

        parser.add_argument("-v", "--varname", type=str, nargs='+',help="enter a valid variable name")
        parser.add_argument("-b", "--branchname", type=str,nargs='+', help="enter a valid branch name")
        parser.add_argument("-u", "--unitout", type=str, help="enter a valid unit")
        parser.add_argument("-pu", "--profileunit", type=str, help="enter a valid profiles unit")
        parser.add_argument("-tu", "--timeunit", type=str, help="enter a valid time unit")
        parser.add_argument(
            "-c",
            "--csv_file",
            type=str,
            help="Path to the CSV file containing variable names, branch names, and pipe names",
        )
        parser.add_argument(
            "-bm",
            "--branch_csv_file",
            type=str,
            help="Path to the CSV file containing  branch names to be sorted",
        )
        args = parser.parse_args()
        args.filepath = args.filepath.replace("\\", "/")
        #args.filepath = [fp.replace("\\", "/") for fp in args.filepath]

        input_matrix = pd.read_csv(args.csv_file)
        # branch_matrix = pd.read_csv(args.branch_csv_file)
        # pplbatchparser = pplBatchParser(args.filepath)
        # trends = pplbatchparser.extract_trends(input_matrix)
        # nodes = pplbatchparser.Join_batch_nodes(input_matrix= input_matrix, branch_matrix=branch_matrix)
        # # print(trends)
        branch_matrix = pd.read_csv(args.branch_csv_file)
        pplparser = pplParser(args.filepath)
        # time_series = pplparser._extract_time_series_data()
        # catalog = pplparser._extract_catalog()
        # # trends = pplparser.extract_trend(var_name=args.varname)
        # profiles = pplparser._extract_branch_profiles(target_branch = args.varname)
        # trends = pplparser.extract_trend(input_matrix= input_matrix)
        nodes = pplparser.extract_trends_join_nodes(input_matrix= input_matrix, branch_matrix=branch_matrix)
        print(nodes)
    # results = pstats.Stats(profile)
    # results.sort_stats(pstats.SortKey.TIME)
    # results.print_stats(20)


import sys
import os
import pandas as pd

sys.path.append("../src/")
import pytest
from src.tpl import tplparser


@pytest.fixture
def command_line_args(request):
    """
    A common fixture which could be passed to all tests in the test script
    """
    args = {}
    args["filepath"] = request.config.getoption("--filepath")
    args["variable"] = request.config.getoption("--variable")
    args["branch"] = request.config.getoption("--branch")
    args["pipe"] = request.config.getoption("--pipe")
    args["index"] = request.config.getoption("--index")
    args["value"] = request.config.getoption("--value")
    return args


def test_invalid_variable_input(command_line_args):
    """
    A test to check if a ValueError is raised if an incorrect variable name is specified
    when the user inputs an extract_trend method
    """
    filepath = command_line_args["filepath"]
    parse = tplparser.tplParser(filepath)
    var = "variable"
    with pytest.raises(ValueError) as excinfo:
        # If ValueError is raised, test passes
        parse.search_catalog(var)
        parse.extract_trend(var)


# def test_valid_variable_input(command_line_args):
#     filepath = command_line_args["filepath"]
#     parse = parser.tplParser(filepath)
#     var = command_line_args["variable"]
#     branch = command_line_args["branch"]
#     pipe = command_line_args["pipe"]
#     parse.search_catalog(var)
#     trend = parse.extract_trend(var, branch, pipe)
#     assert trend is not None


# def test_variable_input(command_line_args):
#     filepath = command_line_args["filepath"]
#     parse = parser.tplParser(filepath)
#     var = command_line_args["variable"]
#     branch = command_line_args["branch"]
#     pipe = command_line_args["pipe"]
#     parse.search_catalog(var)
#     try:
#         trend = parse.extract_trend(var, branch, pipe)
#     except ValueError:
#         with pytest.raises(ValueError):
#             parse.extract_trend(var, branch, pipe)


def test_invalid_branch_input(command_line_args):
    """
    A test to check if a ValueError is raised if an incorrect branch name is specified
    when the user inputs an _extract_branch_profiles method
    """
    parse = tplparser.tplParser(command_line_args["filepath"])
    branch = "branch"
    with pytest.raises(ValueError) as excinfo:
        # If ValueError is raised, test passes
        parse._extract_branch_profiles(branch)


def test_invalid_file_specified(command_line_args):
    """
    A test to check if a file exists
    """
    with pytest.raises(FileNotFoundError) as excinfo:
        parse = tplparser.tplParser(command_line_args["filepath"])



from setuptools import find_packages, find_namespace_packages, setup
import io
import os


# Package Metadata
NAME = "pypontem"
DESCRIPTION = "The one-stop python toolkit for Flow Assurance workflows"
URL = "https://github.com/Pontem-Analytics/pypontem_public/tree/main"

REQUIRES_PYTHON = ">=3.12.3"
VERSION = "1.0.2"
REQUIRED = ["pandas==2.2.0", "pint==0.23", "PyYAML==6.0.2"]
# EXTRAS = {""}

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    # author=AUTHOR,
    # author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(
        where="src", exclude=["tests", "*.tests", "*.tests.*", "tests.*"]
    ),
    # packages=["pypontem", "pypontem/tpl", "pypontem/utils"],
    # package_dir={"pypontem": "src/pypontem"},
    package_dir={"": "src"},
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    # extras_require=EXTRAS,
    include_package_data=True,
    package_data={'': ['utils/*.yaml'],},

    # license='MIT',
    # classifiers=[
    #     # Trove classifiers
    #     # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    #     'License :: OSI Approved :: MIT License',
    #     'Programming Language :: Python',
    #     'Programming Language :: Python :: 3',
    #     'Programming Language :: Python :: 3.6',
    #     'Programming Language :: Python :: Implementation :: CPython',
    #     'Programming Language :: Python :: Implementation :: PyPy'
    # ],
    # $ setup.py publish support.
    # cmdclass={
    #     "upload": UploadCommand,
    # },
)
